r"""Walk a directory, tag its Python files and make a database of the results.

USAGE:
    ```
    paroxython collect [options] DIRECTORY
    ```

OPTIONS:
    ```plain
    -c --cleanup=STR    Transformation to apply to the source code before
                        processing. [default: full]
                        Currently available:
                        • full: remove all blank lines, docstrings and
                                comments (except Paroxython hints).
                        • none: no transformation is applied.
    -e --skip=REGEX     Filter out any file whose name fully matches this
                        regular expression. NB: --glob precedes --skip.
                        [default: (__init__|setup|.*[-_]tests?)\.py]
    -g --glob=PATTERN   The names of the collected files must match this Unix-
                        like glob pattern (not a regular expression). Special
                        syntax: "**/" prefix means “this directory and all
                        subdirectories, recursively”. [default: **/*.py]
    --log               Print a detailed report of the labelling times.
    --no_timestamp      Don't store programs' last modification date.
    -o --output=PATH    The path of the resulting database. If not specified,
                        create a JSON file under the name "DIRECTORY_db.json".
                        Otherwise, use the extension (either ".json", ".sql"
                        or ".sqlite") to decide the format of the tag database.
                        [default: ]
    -t --taxonomy=PATH  The path of a TSV file mapping labels onto taxa. If not
                        specified, use the "taxonomy.tsv" present in DIRECTORY's
                        parent. If absent, use the included default taxonomy:
                        https://github.com/laowantong/paroxython/blob/0.6.1/paroxython/resources/taxonomy.tsv
    ```
"""

from pathlib import Path

from .goodies import print_exit
from .make_db import TagDatabase


def cli_wrapper(args):
    directory = Path(args["DIRECTORY"])
    if not directory.is_dir():
        print_exit(f"no directory at '{directory.absolute()}': aborted.")
    taxonomy = args.get("--taxonomy")
    if taxonomy:
        taxonomy_path = Path(taxonomy)
    else:
        taxonomy_path = directory.parent / "taxonomy.tsv"
        if not taxonomy_path.is_file():
            taxonomy_path = None
    db = TagDatabase(
        directory=directory,
        ignore_timestamps=args["--no_timestamp"],
        cleanup_strategy=args["--cleanup"],  # -> list_programs
        skip_pattern=args["--skip"],  # -> list_programs
        glob_pattern=args["--glob"],  # -> list_programs
        print_performances=args["--log"],  # -> labelled_programs
        taxonomy_path=taxonomy_path,  # -> Taxonomy
    )
    if not args["--output"]:
        db.write_json()
    elif args["--output"].endswith(".json"):
        db.write_json(Path(args["--output"]))
    elif args["--output"].endswith((".sqlite", ".sql")):
        db.write_sqlite(Path(args["--output"]))



r"""
Translate labels into taxa.

## Description

This step comes after `paroxython.label_programs` has tagged a program with the **labels** specified
in `spec.md`. Its purpose is to convert these labels into **taxa** based on the rules defined (by
default) in `taxonomy.tsv`.

## Example

Let's return to the Jupyter notebook's cell given in the introduction:

```python
1   %%paroxython labels
2   def fibonacci(n):
3       result = []
4       (a, b) = (0, 1)
5       while a < n:
6           result.append(a)
7           (a, b) = (b, a + b)
8       return result
```

Note that the magic cell-command `%%paroxython` is now called with a `labels` argument. This
produces the following output (limited for brevity to its first four rows):

| Label | Lines |
|:--|:--|
| `addition_operator` | 7 |
| `assignment` | 3, 4, 7 |
| `assignment_lhs_identifier:a` | 4, 7 |
| `assignment_lhs_identifier:b` | 4, 7 |

These labels are low-level tags, intended for internal use only. They still need to be translated
into the following taxa, which are the only tags of interest to the end user. For instance,
`"addition_operator"` will be translated into `"operator/arithmetic/addition"` (note that, unlike
a label, a taxon may feature one or several slashes, which indicate the nesting of notions). More
examples are given in the dedicated [section](user_manual/index.html#taxonomy) of the user manual.

In fact, both labels and taxa are couples:

- Label: `("addition_operator", [(7, 7, "...")])`.
- Taxon: `("operator/arithmetic/addition", {(7, 7, "..."): 1})`.

The second member is slightly more complicated than a list of line numbers:

- For a label, it is a list of triples consisting of the start and end of the spanning lines,
and a third member, the path (here denoted by the ellipsis), which identifies unambiguously the
start of the span.
- For a taxon, it is a **bag**, i.e. a dictionary associating such triples with the count of their
occurrences in the program.

The paths being left untouched by the conversion, we will omit them from now on.

### Deduplication

- Labels:
    - `("assignment", [(3, 3), (4, 4), (7, 7)])`.
    - `("single_assignment:result", [(3, 3)])`.
    - `("parallel_assignment", [(4, 4), (7, 7)])`.
    - `("slide", [(7, 7)])`.
- Taxa:
    - `("var/assignment/explicit/single", {(3, 3): 1})`.
    - `("var/assignment/explicit/parallel", {(4, 4): 1})`.
    - `("var/assignment/explicit/parallel/slide", {(7, 7): 1})`.

The example program features three distinct assignments. They can be categorized into:

- A single assignment on line 3  (`result = []`).
- A parallel assignment on line 4 (`(a, b) = (0, 1)`).
- A parallel assignment where a couple is updated by “sliding” the value of its second member to the
first position, on line 7 (`(a, b) = (b, a + b)`).

As you can see, labelling all these produces a lot of span redundancies. For instance, the
assignment of line 7 is labelled three times, as `"assignment"`, `"parallel_assignment"` and
`"slide"`. Since the labels are not structured, this is the best that we can do.

Enter the taxa, whose production is triggered by the following taxonomy rows:

Taxa (replacement patterns)    | Labels (search patterns)
:------------------------------|:-----------------------
`var/assignment/explicit` | `assignment\b.*`
`var/assignment/explicit/parallel` | `parallel_assignment:\d+`
`var/assignment/explicit/parallel/slide` | `slide`
`var/assignment/explicit/single` | `single_assignment:.+`

Initially, the conversion produces as many redundant spans as for the labels:

- Taxa:
    - `("var/assignment/explicit", {(3, 3): 1, (4, 4): 1, (7, 7): 1})`.
    - `("var/assignment/explicit/single", {(3, 3): 1})`.
    - `("var/assignment/explicit/parallel", {(4, 4): 1, (7, 7): 1})`.
    - `("var/assignment/explicit/parallel/slide", {(7, 7): 1})`.

However, because of the tree-like structuring of taxa, it is possible to remove those occurrences
that are subject to further characterization (e.g., `var/assignment/explicit/parallel` on line 7 is
further characterized as `var/assignment/explicit/parallel/slide`), resulting in the list of taxa
given at the start of the section.

Note that the removal of spans has finally led to the elimination of the overly generic taxon
`var/assignment/explicit`.
"""


from collections import Counter, defaultdict
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple
from typing import Pattern as RegexPattern
from os.path import commonpath, dirname
from functools import lru_cache

import regex  # type: ignore

from .user_types import (
    LabelName,
    LabelPattern,
    Labels,
    Taxon,
    TaxonName,
    TaxonNames,
    TaxonPattern,
    Taxa,
    TaxaSpans,
)

__pdoc__ = {
    "Taxonomy.__init__": True,
    "Taxonomy": "",
}


class Taxonomy:
    def __init__(
        self,
        taxonomy_path: Optional[Path] = None,
        **kwargs,
    ) -> None:
        r"""Read and pre-process the taxonomy specifications.

        Args:
            taxonomy_path (Optional[Path], optional): The path of a two-columns TSV file
                associating label (search) patterns with taxon (replacement) patterns. For better
                readability, the taxa are listed on the first column, and the corresponding labels
                (sometimes very long) on the second column. If not specified, the
                [default taxonomy](https://repo/paroxython/resources/taxonomy.tsv) is used.
                Defaults to `None`.

        Description:
            1. Read the taxonomy, falling back to the provided default `taxonomy.tsv`.
            2. Drop the part starting with the string `"-- EOF"`, if any (this pre-processing
            allows you to leave notes or drafts at the end of the file).
            3. Distribute the remaining rows into two different accumulators, depending on the
            “literality” of the label pattern:
                - `self.literal_labels` is a dictionary associating every **literal** label pattern,
                that it to say, a label name, with a list of taxon names. Most of these lists are
                singletons, but one can imagine that a unique label name could be associated with
                multiple taxon names. As a dictionary, it allows O(1) lookups of any given label
                pattern.
                - `self.compiled_labels` is a list of couples, of which the first member is a
                **compiled** label pattern, and the second member is a replacement pattern which may
                contain backreferences to one or several capture groups of the label pattern. For
                instance, it can include the couple `(regex.compile(r"nested_index:(\d+)$"),
                r"subscript/index/shape/\1")`, which will be used to “translate” any occurrence of
                the label name `"nested_index:3"` into the taxon name `"subscript/index/shape/3"`.
                As an associative list, it only allows O(n) lookups a given label pattern.
        """
        taxonomy_path = taxonomy_path or Path(dirname(__file__)) / "resources" / "taxonomy.tsv"
        tsv = taxonomy_path.read_text().partition("-- EOF")[0].strip()
        self.literal_labels: Dict[LabelName, TaxonNames] = defaultdict(list)
        self.compiled_labels: List[Tuple[RegexPattern, TaxonPattern]] = []
        for line in sorted(tsv.split("\n")[1:]):
            (taxon_value, label_value, *_) = line.strip().split(maxsplit=2)
            (taxon_pattern, label_pattern) = (TaxonPattern(taxon_value), LabelPattern(label_value))
            if is_literal(label_pattern):
                self.literal_labels[LabelName(label_pattern)].append(TaxonName(taxon_pattern))
            else:
                self.compiled_labels.append((regex.compile(f"{label_pattern}$"), taxon_pattern))
                # note: "$" is necessary: regex.fullmatch() has no regex.fullsub() counterpart

    @lru_cache(maxsize=None)
    def get_taxon_name_list(
        self, label_name: LabelName, looks_like_a_taxon: Callable = regex.compile(r"^\w+/.+$").match
    ) -> TaxonNames:
        r"""Translate a label name into a list of taxon names.

        Description:
            First of all, when a label looks like a taxon, it is returned without further ado.

            For the rest, most of the work was done during the initialization, by constructing the
            map `self.literal_labels` and the list `self.compiled_labels`. The given label name is
            first looked up in the map, then matched successively against every regular expression
            stored in the list.

            The second step may be useful even if the first one has already produced one or several
            translations. For instance, take a taxonomy defining the following associations:

            Taxa (replacement patterns)         | Labels (search patterns)
            :-----------------------------------|:-----------------------
            `call/subroutine/builtin/casting/\1`  | `free_call:(list|dict)`
            `type/sequence/list`                | `free_call:list`

            Suppose now that we pass the label `"free_call:list"`. The lookup in the map produces
            a first translation: `"type/sequence/list"`. However, the linear search in the list
            produces a second translation, namely: `"call/subroutine/builtin/casting/list"`.

            >>> a_taxonomy_instance.get_taxon_name_list("free_call:list")
            [
                "type/sequence/list",
                "call/subroutine/builtin/casting/list",
            ]

            Therefore, the complexity of this function is linear in every case. However, its
            [memoization](https://en.wikipedia.org/wiki/Memoization) avoids having to translate the
            same label more than one time
            (cf. [lru_cache](https://docs.python.org/3/library/functools.html#functools.lru_cache)).
        """
        if looks_like_a_taxon(label_name):
            return [TaxonName(label_name)]
        result: TaxonNames = self.literal_labels.get(label_name, [])
        for (label_regex_pattern, taxon_pattern) in self.compiled_labels:
            if label_regex_pattern.match(label_name):
                result.append(label_regex_pattern.sub(taxon_pattern, label_name))
        return result

    def to_taxa(self, labels: Labels) -> Taxa:
        """Translate a list of labels to a list of taxa.

        Args:
            labels (Labels): A list of labels, each label consisting in a name and a **list** of
                spans.

        Returns:
            Taxa: A sorted list of taxa, each taxon consisting in a name and a **bag** of spans.

        Description:
            This simple method is the sole entry point of the object. It takes a list of labels,
            accumulates their translations into taxa (cf. `Taxonomy.get_taxon_name_list`) and
            returns them deduplicated (cf. `deduplicated_taxa`).
        """
        acc: TaxaSpans = defaultdict(Counter)
        for (label_name, spans) in labels:
            for taxon_name in self.get_taxon_name_list(label_name):
                acc[taxon_name].update(spans)
        taxa = [Taxon(name, spans) for (name, spans) in sorted(acc.items())]
        return deduplicated_taxa(taxa)


def is_literal(label_pattern: LabelPattern) -> bool:
    r"""Tell whether a given regular expression pattern is literal, disregarding the dots.

    Description:
        All label patterns of the taxonomy TSV file (second column) will be tested successively. For
        instance, the label pattern `"external_free_call:print"` is literal, and then matches only
        itself. Conversely, the label pattern `"internal_free_call:[[:upper:]].*"` contains some
        non-literal characters such as `"["` and `"*"`, and then must be compiled into a regular
        expression. A pattern is considered to be literal if it is equal to its escaped form, **not
        taking the dots into account**.

    ..note::
        In case not considering the dot as a metacharacter is a  problem, it is always possible to
        force the interpretation of a pattern as non-literal by enclosing it in parentheses.
    """
    return label_pattern.replace(".", "\\.") == regex.escape(label_pattern)


def deduplicated_taxa(taxa: Taxa) -> Taxa:
    """Remove from the given taxa those that are counted multiple times as prefixes of another ones.

    Description:
        Consider the following program:
        ```python
        1   while n != 1:
        2       if n % 2 == 0:
        3           n = n // 2
        4       else:
        5           n = 3 * n + 1
        ```
        It features exactly one literal integer on lines 1 and 3, and exactly two on lines 2 and 5.
        Moreover, one of the literals of line 2 is zero. Therefore, the program is originally tagged
        with the following two couples (among others):
        ```python
        [
            ...
            ("type/number/integer/literal",
                {
                    (1, 1): 1,
                    (2, 2): 2,
                    (3, 3): 1,
                    (5, 5): 2,
                }
            ),
            ("type/number/integer/literal/zero", {(2, 2): 1}),
            ...
        ]
        ```
        ..tip::
            In the dictionary above, `(5, 5): 2` means that the taxon `"type/number/integer/literal"`
            has 2 distinct occurrences spanning the lines 5 to 5 (namely, `3` and `1`). Such a
            “counter” dictionary is called a _bag_, or _multiset_.

        Now, every occurrence of a given taxon **is** an occurrence of any of its prefix. For
        instance, an occurrence of the literal integer `0` is obviously an occurrence of a literal
        integer. Currently, it is counted twice. The purpose of this function is to remove such
        redundant counts. The longest taxon will be kept, and its prefix count decreased. Finally,
        the following deduplicated list will be returned:
        ```python
        [
            ...
            ("type/number/integer/literal",
                {
                    (1, 1): 1,
                    (2, 2): 1, # decremented
                    (3, 3): 1,
                    (5, 5): 2,
                }
            ),
            ("type/number/integer/literal/zero", {(2, 2): 1}),
            ...
        ]
        ```
        Decrementing a span count may result in the deletion of a taxon. For instance, on the same
        program:
        ```python
        [
            ...
            ("flow/loop", {(1, 5): 1},
            ("flow/loop/exit/late", {(1, 5): 1},
            ("flow/loop/while", {(1, 5): 1},
            ...
        ]
        ```
        … becomes after deduplication:
        ```python
        [
            ...
            ("flow/loop/exit/late", {(1, 5): 1},
            ("flow/loop/while", {(1, 5): 1},
            ...
        ]
        ```

    Args:
        taxa (Taxa): A list of couples, sorted lexicographically and consisting of a taxon name and
            a bag (multiset) of its spans.

    Returns:
        Taxa: The same list, deduplicated.
    """
    if len(taxa) < 2:
        return taxa
    for (i, (name, spans)) in enumerate(taxa[1:], 1):
        for (previous_name, previous_spans) in reversed(taxa[:i]):
            # Since `"/"` happens to be a separator in both taxa and Unix-paths, `os.commonpath()`
            # comes in handy for extracting the common prefix of two taxa:
            common_prefix = commonpath((name, previous_name))
            if not common_prefix:
                break
            if previous_name == common_prefix:
                difference = spans - previous_spans
                previous_spans.subtract(spans)
                spans = difference
    result = []
    for (name, spans) in taxa:
        spans += Counter()  # suppress items whose count <= 0
        if spans:  # if any item remains in the bag
            result.append(Taxon(name, spans))
    return result


if __name__ == "__main__":
    from .goodies import couple_to_string
    from .label_programs import labelled_programs

    taxonomy = Taxonomy()
    for program in labelled_programs(Path("examples/simple/programs")):
        taxa = taxonomy.to_taxa(program.labels)
        if not taxa:
            continue
        width = min(40, max(len(" ".join(map(str, taxon.spans))) for taxon in taxa))
        for (name, spans) in taxa:
            span_string = " ".join(map(couple_to_string, sorted(set(spans))))
            print(f"{span_string:>{width}}\t{name}")



"""Process an input string to interpret it as a known predicate."""

from typing import Callable, Tuple

import regex  # type: ignore

from .compare_spans import compare_spans
from .goodies import print_warning, print_fail
from .user_types import Predicate


def normalize_predicate(
    predicate: str,
    search_not_1: Callable = regex.compile(r"not\s+").search,
    search_not_2: Callable = regex.compile(r"\s+not").search,
    sub_is: Callable = regex.compile(r" is\b|\bis ").sub,
    sub_forbidden_chars: Callable = regex.compile(r"[^xy<=≤]").sub,
    sub_identity: Callable = regex.compile(r"^(x=y|y=x)$").sub,
    sub_one_x: Callable = regex.compile(r"^([^x]*)x([^x]*)$").sub,
    sub_one_y: Callable = regex.compile(r"^([^y]*)y([^y]*)$").sub,
) -> Tuple[Predicate, bool]:
    """Ensure that the given predicate is correct or can be salvaged.

    Description:
        The system first determines whether the predicate is negated or not, which can be expressed
        either by an exclamation mark prefix or the word “not”. In both cases, the negation marker
        is extracted from the given string.

        The resulting string is then normalized, i.e. replaced by a key of the dictionary
        `compare_spans` defined in `paroxython.compare_spans`.
        .. include:: ../docs/md/pipeline_documentation.md
            :start-after: user input.
            :end-before: ### Positive semantic triples

    Examples:
        Browse
        [`test_normalize_predicate.py`](https://repo/tests/test_normalize_predicate.py)
        on GitHub.

    Args:
        predicate (str): Plain string to be interpreted as a predicate.
        search_not_1 (Callable, optional): Function searching `"not "`.
        search_not_2 (Callable, optional): Function searching `" not"`.
        sub_is (Callable, optional): Function replacing `" is"` or `"is "`.
        sub_forbidden_chars (Callable, optional): Function replacing all characters not in `"xy<=≤"`.
        sub_identity (Callable, optional): Function replacing `"x=y"` or `"y=x"`.
        sub_one_x (Callable, optional): Function replacing isolated`"x"`.
        sub_one_y (Callable, optional): Function replacing isolated`"y"`.

    Note:
        [No default argument to be explicitly provided.](developer_manual/index.html#default-argument-trick)

    Raises:
        ValueError: Raised when the input string is malformed beyond recognition.

    Returns:
        Tuple[Predicate, bool]: A tuple whose first member is a function (more precisely, a value
            of the dictionary `compare_spans`), and second member is a boolean indicating whether
            the result of this predicate must be negated.
    """
    predicate = predicate.lower().strip()
    negated = True
    if predicate.startswith("!"):
        predicate = predicate[1:]
    elif search_not_1(predicate):
        predicate = predicate.replace("not ", "")
    elif search_not_2(predicate):
        predicate = predicate.replace(" not", "")
    else:
        negated = False
    original = predicate = predicate.strip()
    predicate = sub_is("", predicate)  # Suppress "is" when it is followed or preceded by a space.
    if predicate not in compare_spans:
        predicate = predicate.replace("<=", "≤").replace("==", "=")  # Convert usual operators.
        predicate = sub_forbidden_chars("", predicate)  # Ignore all other characters.
        predicate = sub_identity("x=y≤x=y", predicate)  # Treat the special case of identity.
        predicate = sub_one_x(r"\1x≤x\2", predicate)  # If there is only one x, expand it into x≤x.
        predicate = sub_one_y(r"\1y≤y\2", predicate)  # If there is only one y, expand it into y≤y.
        if original != predicate:
            print_warning(f"predicate '{original}' normalized into '{predicate}'.")
        if predicate not in compare_spans:
            print_fail(f"Malformed predicate '{predicate}' in the pipeline.")
    return (compare_spans[predicate], negated)



"""A grab bag of various generic functions and factories."""

import sys
from collections import defaultdict
from math import log2
from textwrap import wrap
from typing import Callable, Dict, Tuple, Union
from unicodedata import normalize

import regex  # type: ignore

from .user_types import Span

# ANSI color codes
OK = "\033[92m"
WARNING = "\033[1m\033[38;5;166m"
FAIL = "\033[1m\033[91m"
RESET = "\033[0m"


def title_to_slug_factory() -> Callable:
    """Return a function mapping a string to the appropriate slug, optionally disambiguated.

    Returns:
        Callable[[str, Optional[bool]]]: A function taking a string and an optional boolean
            `deduplicate` (default: `False`). If `True`, when two or more strings produce the same
            slug, a numeric suffix is added to the second one (`"-1"`), the third one (`"-2"`), and
            so on.

    Examples:
        See [`test_goodies.py`](https://repo/tests/test_goodies.py).
    """
    slug_counts: Dict[str, int] = defaultdict(int)
    cache: Dict[str, str] = {}

    def title_to_slug(title, deduplicate=False):
        title = title.strip()
        if title not in cache:
            slug = normalize("NFD", title.lower()).encode("ASCII", "ignore").decode("ASCII")
            slug = slug.replace(" ", "-")
            slug = regex.sub(r"[^\w-]", "", slug)
            cache[title] = slug
        if deduplicate:
            slug = cache[title]
            slug_counts[slug] += 1
            slug = f"{slug}-{slug_counts[slug] - 1}"
            slug_counts[slug] += 1
            slug = slug.rstrip("-0")
            return slug
        else:
            return cache[title]

    return title_to_slug


def add_line_numbers(source: str) -> str:
    """Return a numbered version of the given source code. Result readable up to 999 lines."""
    if not source:
        return ""
    return "\n".join(f"{n: <4}{line}" for (n, line) in enumerate(source.split("\n"), 1))


def enumeration_to_txt_factory(
    width: int = 20,
    default_string: str = "",
    sep: str = "<br>",
    template: str = "<details><summary>{summary}</summary>{details}</details>",
    initial_indent: int = 3,  # take into account the details marker
) -> Callable[[str], str]:
    """Return a function formatting an enumeration on a given column width.

    This function is used by `recommend_programs.get_markdown()` to limit the width of the last
    column of the Markdown tables, as in the following example:

    | Cost  | Taxon | Location |
    |:---|:---|:---|
    | 0.9375 | `type/number/integer/literal` | <details><summary>1, 1, 1, 1, </summary>1, 1, 2, 2, 2, <br>2, 2, 3, 3, 3</details> |

    The enumeration, given as a string, is wrapped onto one or several lines of at most `width`
    characters. The first line and the remaining ones are formatted separately according to the
    given `template`.

    Args:
        width (int, optional): Column width.
            Defaults to `20`.
        default_string (str, optional): String to return when the enumeration is empty.
            Defaults to `""`.
        sep (str, optional): Separator between wrapped lines.
            Defaults to `"<br>"`.
        template (str, optional): A formatting string on the first line (referred as `summary`)
            and the remaining ones (referred as `details`).
            Defaults to `"<details><summary>{summary}</summary>{details}</details>"`.
        initial_indent (str, optional): Approximate space-width of the detail marker (e.g.,
            `"▶︎ "`).
            Defaults to `3`.

    Returns:
        Callable[[str], str]: Function returning a wrapped version of a given string.

    Example:
        >>> enumeration_to_txt = enumeration_to_txt_factory(7)  # create the function
        >>> enumeration_to_txt("1, 2, 3, 4, 5-6, 7, 8, 9")      # use it
        "<details><summary>1,</summary>2, 3,<br>4, 5-6,<br>7, 8, 9</details>"
    """

    def enumeration_to_txt(s: str) -> str:
        if not s:
            return default_string
        if len(s) <= width:
            return s
        lines = wrap(s, width, initial_indent=" " * initial_indent)
        summary = lines[0][initial_indent:]
        details = sep.join(lines[1:])
        return template.format(summary=summary, details=details)

    return enumeration_to_txt


def cost_bucket(cost: int) -> str:
    r"""Return an interval including a given positive number.

    The intervals are predefined as \([0, 0], ]0, 0.25[, [2^{i-2}, 2^{i-1}[, \dots\) for all
    natural numbers \(i\). The result is returned as a string ready to be included in the
    recommendation report.

    Examples:
        >>> cost_bucket(0)
        "0"
        >>> cost_bucket(0.75)
        "in [0.5, 1["
    """
    if cost == 0:
        return "0"
    if cost < 0.25:
        return "in ]0, 0.25["
    if cost < 0.5:
        return "in [0.25, 0.5["
    if cost < 1:
        return "in [0.5, 1["
    upper = 2 ** int(log2(cost))
    lower = 2 ** int(log2(cost) + 1)
    return f"in [{upper}, {lower}["


def couple_to_string(couple: Union[Span, Tuple[int, int]]) -> str:
    """Return a deduplicated string representation of the given couple or span.

    Examples:
        >>> couple_to_string((12, 15))
        "12-15"
        >>> couple_to_string((12, 12))
        "12"
        >>> couple_to_string(Span(12, 15))
        "12-15"
    """
    return f"{couple[0]}" + ("" if couple[0] == couple[1] else f"-{couple[1]}")


def print_success(message: str):
    """Print the given message of success."""
    print(f"{OK}{message}{RESET}")


def print_warning(message: str):
    """Print the given message on `sys.stderr`."""
    print(f"{WARNING}Warning: {message}{RESET}", file=sys.stderr)


def print_fail(message: str):
    """Print the given message on `sys.stderr` and raise a ValueError."""
    print(f"{FAIL}Error: {message}{RESET}", file=sys.stderr)
    raise ValueError(message)


def print_exit(message: str):
    """Print the given message on `sys.stderr` and exit."""
    sys.exit(f"{FAIL}Error: {message}{RESET}")



from typing import Counter as Bag
from typing import Dict, List, NamedTuple, NewType, Set, Tuple, Union, Callable

from typing_extensions import TypedDict, Literal  # Python 3.8: import directly from typing

# fmt: off


# --------------------------------------------------------------------------------------------------
# Various types
# --------------------------------------------------------------------------------------------------

Source = NewType("Source", str)  # Source code of the user's programs.
Query = NewType("Query", str)  # SQL specification of a feature, as defined in `spec.md`.

class Span(NamedTuple):  # Location of a given feature.
    start: int  # First line of the feature.
    end: int  # Last line of the feature.
    path: str = ""  # Identifier of the beginning of the feature in the AST.


# --------------------------------------------------------------------------------------------------
# Labels
# --------------------------------------------------------------------------------------------------

LabelName = NewType("LabelName", str)  # Name of a feature, as defined in `spec.md`.
LabelNames = List[LabelName]

class Label(NamedTuple):
    name: LabelName
    spans: List[Span]

Labels = List[Label]
LabelsSpans = Dict[LabelName, List[Span]]

LabelPattern = NewType("LabelPattern", str)  # Member of the second row of `taxonomy.tsv`.


# --------------------------------------------------------------------------------------------------
# Taxa
# --------------------------------------------------------------------------------------------------

TaxonName = NewType("TaxonName", str) # Path from a root to a node in the taxonomy.
TaxonNames = List[TaxonName]
TaxonNameSet = Set[TaxonName]

class Taxon(NamedTuple):
    name: TaxonName
    spans: Bag[Span]  # For a given taxon, associate each span to its number of occurrences.
                      # For instance, if there are three integer literals on line 42, the
                      # corresponding entry is: (42, 42, '...'): 3.

Taxa = List[Taxon]
TaxaSpans = Dict[TaxonName, Bag[Span]]

TaxonPattern = NewType("TaxonPattern", str) # Member of the first row of `taxonomy.tsv`.
TaxonPatterns = List[TaxonPattern]


# --------------------------------------------------------------------------------------------------
# Programs
# --------------------------------------------------------------------------------------------------

ProgramPath = NewType("ProgramPath", str)  # relative path of a program in the directory to search
ProgramPaths = List[ProgramPath]
ProgramPathSet = Set[ProgramPath]

class Program(NamedTuple):  # All the information pertaining to a program.
    labels: Labels
    taxa: Taxa
    addition: LabelsSpans  # Features scheduled for addition.
    deletion: LabelsSpans  # Features scheduled for deletion.
    path: ProgramPath = ProgramPath("")
    source: Source = Source("")

Programs = List[Program]
ProgramTaxa = Dict[ProgramPath, Taxa]


# --------------------------------------------------------------------------------------------------
# Serialization-ready types used for the JSON tag database
# --------------------------------------------------------------------------------------------------

PoorSpan = Tuple[int, int]  # A span, deprived from its path.
LabelsPoorSpans = Dict[LabelName, List[PoorSpan]]
TaxaPoorSpans = Dict[TaxonName, List[PoorSpan]]

class ProgramRecord(TypedDict):
    timestamp: str  # Date of last modification: currently unused.
    source: Source
    labels: LabelsPoorSpans
    taxa: TaxaPoorSpans

ProgramInfos = Dict[ProgramPath, ProgramRecord]
LabelInfos = Dict[LabelName, ProgramPaths]
TaxonInfos = Dict[TaxonName, ProgramPaths]
ProgramToPrograms = Dict[ProgramPath, ProgramPaths]  # For import-export stuff.

class JsonTagDatabase(TypedDict):  # Schema of the JSON version of a tag database.
    programs: ProgramInfos
    labels: LabelInfos
    taxa: TaxonInfos
    importations: ProgramToPrograms
    exportations: ProgramToPrograms


# --------------------------------------------------------------------------------------------------
# Pipeline dictionary
# --------------------------------------------------------------------------------------------------

Operation = NewType("Operation", str) # "exclude", "include", "include all", "include any", etc.

Criterion = Union[str, Tuple[str, str, str]] # Either a pattern or a semantic triple.

class Command(TypedDict):
    operation: Operation
    data: Union[str, List[Criterion]]  # Either a shell command or a list of criteria.

Predicate = Callable[[PoorSpan, PoorSpan], bool] # A value of the dictionary `compare_spans`.


# --------------------------------------------------------------------------------------------------
# Recommendations
# --------------------------------------------------------------------------------------------------

AssessedPrograms = List[Tuple[float, ProgramPath]] # Associative array between costs and programs
AssessmentStrategy = Literal["zeno", "linear"]

# fmt:on



"""
Bring together everything relative to the pre-processing of a given source code.

## Cleaning

A useful, albeit not essential, step is to cleanse the code of all its algorithmically irrelevant
features (blank lines, comments, docstrings, etc.). Lots of examples are provided in the `Cleanup`
class documentation below.

## Manual hints: preservation, centrifugation and scheduling

Certain comments intentionally survive the cleaning step, the so-called “manual hints”, introduced
[here](user_manual/index.html#manual-hints).

- If they appear at the end of a nonempty line of code, they are preserved as they are.
- If they extend over an entire line (excluding the initial blanks), they are considered to be
  featured by the entire program. They are replaced by two hints, one put at the beginning and
  the other at the end of the source code, thus delimiting the totality of the program. This
  operation is called “centrifugation”.

In either case, these hints were manually placed by the user to modify the behavior of Paroxython,
by asking it either to add the taxa that it missed, or to delete those that it unduly produced. This
is at this step that they are scheduled for addition or deletion (respectively). For the moment, the
information is just stored: `paroxython.parse_program.ProgramParser` will process it when required.
"""

from collections import defaultdict
from token import DEDENT, INDENT, NEWLINE, STRING
from tokenize import COMMENT, NL, generate_tokens
from typing import Callable, Dict, List, Tuple, Set

import regex  # type: ignore

from .goodies import print_fail
from .user_types import LabelName, LabelsSpans, Source, Span

HINT_COMMENT = "# paroxython:"


class Cleanup:
    def __init__(self, cleanup_strategy: str):
        """Collection of cleaning static methods to be applied to a source code.

        Args:
            cleanup_strategy (str): The name of the transformation. Currently, any other name than
                `"full"` (corresponding to `full_cleaning()` comes down to the identity function).

        Returns:
            Callable[[Source], Source]: A function that will clean up a given source code.
        """
        self.run = lambda source: source
        if cleanup_strategy == "full":
            self.run = Cleanup.full_cleaning

    @staticmethod
    def full_cleaning(source: Source) -> Source:
        r'''Remove as much noise (comments, etc.) as possible in the given source code.

        Description:
            1. Suppress initial comments.
            2. Suppress `if __name__ == '__main__'` part.
            3. Suppress `sys.path` injections (frequently used to simplify `import` behavior).
            4. Replace tabulations with spaces.
            5. Normalize Paroxython hint comments.
            6. Replace docstrings with `pass` statements (most of which being suppressed at the end
               of the process).
            7. Suppress empty or blank lines.
            8. Suppress useless `pass` statements.

        Examples:
            - Empty or blank lines are suppressed.
            ```python
            foo = bar


            bar = buzz

            ```
            ```python
            foo = bar
            bar = buzz
            ```
            - Normal inline comments are suppressed.
            ```python
            foo = bar # lorem ipsum
            fizz = [
                "foo", # lorem
                "bar", # ipsum
            ]
            ```
            ```python
            foo = bar
            fizz = [
                "foo",
                "bar",
            ]
            ```
            - Commented lines are suppressed.
            ```python
            # lorem ipsum
            # dolor amet
            def foo(bar):
                # lorem
                # ipsum
                if foo == bar:
                    # lorem
                    return []
                # lorem
            # ipsum
            ```
            ```python
            def foo(bar):
                if foo == bar:
                    return []
            ```
            - Shebang is suppressed.
            ```python
            #!/usr/bin/env python
            foobar()
            ```
            ```python
            foobar()
            ```
            - Encoding declaration is suppressed.
            ```python
            # coding=utf8
            foobar()
            ```
            ```python
            foobar()
            ```
            - Docstrings of classes and functions are suppressed.
            ```python
            class Foo:
                """Lorem.

                Ipsum dolor.
                """
                def foo(self, bar):
                    """Lorem ipsum."""
                    return 1
            def bar():
                """Ipsum dolor"""
                return 2
            ```
            ```python
            class Foo:
                def foo(self, bar):
                    return 1
            def bar():
                return 2
            ```
            - Docstrings of modules are suppressed.
            ```python
            """Lorem.

            Ipsum Dolor."""

            """Lorem ipsum dolor sic amet."""

            a = """consectetur
            adipiscing
            elit.
            """
            ```
            ```python
            a = """consectetur
            adipiscing
            elit.
            """
            ```
            - A `pass` statement is added to empty classes if needed.
            ```python
            class Foo:
                """Lorem.

                Ipsum dolor.
                """
            ```
            ```python
            class Foo:
                pass
            ```
            - Useless `pass` statements are suppressed.
            ```python
            foo()
            pass
            if bar():
                pass
            else:
                pass
                foobar()
                pass
            ```
            ```python
            foo()
            if bar():
                pass
            else:
                foobar()
                pass
            ```
            - Mixes of comments and docstrings are suppressed.
            ```python
            # Lorem
            # Ipsum
            """Dolor."""
            foobar()
            ```
            ```python
            foobar()
            ```
            - Paroxython's label hints are **preserved** and **normalized**.
            ```python
            foo = bar #   Paroxython   :   hint_1   hint_2
            #paroxython: hint_3
            fizz = [
                "foo", # lorem
                "bar", # paroxython: hint_4
            ]
            ```
            ```python
            foo = bar # paroxython: hint_1   hint_2
            # paroxython: hint_3
            fizz = [
                "foo",
                "bar", # paroxython: hint_4
            ]
            ```
            - `if __name__ == "__main__":` part is suppressed.
            ```python
            foo = bar
            if __name__ == "__main__":
                bar = foo
            ```
            ```python
            foo = bar
            ```
            - `sys.path` injections are suppressed.
            ```python
            __import__("sys").path[0:0] = ["programs"]
            (28433 * 2**7830457 + 1) % 10**10
            ```
            ```python
            (28433 * 2**7830457 + 1) % 10**10
            ```

            All examples above automatically extracted from
            [test_cleanup_source.py](https://repo/tests/test_cleanup_source.py).

        .. warning::
            Useless pass statements are preserved when they are not followed by a line with a same
            indentation level.
        '''
        result = []
        previous_token = INDENT
        (previous_end_row, previous_end_col) = (-1, 0)
        text = str(source)
        text = Cleanup.suppress_first_comments(text)
        text = Cleanup.suppress_main_guard(text)
        text = Cleanup.suppress_sys_path_injection(text)
        text = text.replace("\t", "    ")
        lines = iter(text.split("\n"))
        for token_info in generate_tokens(lambda: next(lines) + "\n"):
            (token, string, (start_row, start_col), (end_row, end_col), _) = token_info
            if start_row > previous_end_row:
                previous_end_col = 0
            result.append(" " * max(0, start_col - previous_end_col))
            if token == COMMENT:
                (string, n) = Cleanup.normalize_paroxython_comments(string)
                if n == 0:
                    continue
                result.append(string)
            elif token == STRING and previous_token in (INDENT, DEDENT, NEWLINE):
                result.append("pass\n")  # replace the docstring by a pass statement
            else:
                result.append(string)
            if (previous_token, token) == (NEWLINE, NL):
                previous_token = NEWLINE
            else:
                previous_token = token
            (previous_end_row, previous_end_col) = (end_row, end_col)
        text = "".join(result).strip()
        text = Cleanup.suppress_blank_lines(text)
        text = Cleanup.suppress_useless_pass_statements(text)
        return Source(text)

    @staticmethod
    def suppress_first_comments(
        source: str, sub: Callable = regex.compile(r"\A(#.*\n)*").sub
    ) -> str:
        """Replace all comments placed on the first lines of the source code.

        Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
        """
        return sub("", source)

    @staticmethod
    def suppress_main_guard(
        source: str, sub: Callable = regex.compile(r"(?ms)^if +__name__ *== *.__main__. *:.+").sub
    ) -> str:
        """Suppress `if __name__ == '__main__'` part.

        Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
        """
        return sub("", source)

    @staticmethod
    def suppress_sys_path_injection(
        source: str,
        sub: Callable = regex.compile(r'(?m)^__import__\("sys"\)\.path\[0:0\] = .+\n').sub,
    ) -> str:
        """Suppress lines starting with `__import__("sys").path[0:0] = `.

        Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
        """
        return sub("", source)

    @staticmethod
    def normalize_paroxython_comments(
        source: str,
        subn: Callable = regex.compile(
            r"\s*".join(regex.split(r"(\w+)", HINT_COMMENT.replace(" ", ""))) + r"\s*",
            regex.IGNORECASE,
        ).subn,
    ) -> Tuple[str, int]:
        r"""Replace and count all the hint comment strings, made space- and case-insensitive.

        For instance, if `HINT_COMMENT` is `"# paroxython:"` (its default value), the actual search
        pattern will be `r"(?i)#\s*paroxython\s*:\s*"`.

        Argument `subn` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
        """
        return subn(f"{HINT_COMMENT} ", source)

    @staticmethod
    def suppress_blank_lines(source: str, sub: Callable = regex.compile(r"\s*\n").sub) -> str:
        """Suppress all empty or blank lines in the given source.

        Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
        """
        return sub("\n", source)

    @staticmethod
    def suppress_useless_pass_statements(
        source: str,
        sub: Callable = regex.compile(r"(?m)^( *)pass\n\1(?!\s)").sub,
    ) -> str:
        """Suppress all `pass` statements followed by a line with a same level of indentation.

        Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
        """
        return sub(r"\1", source)


def centrifugate_hints(
    source: Source, match_isolated_hints: Callable = regex.compile(fr"\s*{HINT_COMMENT} (.+)").match
) -> Source:
    """Transform the isolated hints into all-encompassing hints.

    Description:
        When a hint is isolated on its own line, it is considered to extend to the entire program.
        These hints are moved to both ends of the source (centrifugated), and made open on the
        first line and close on the last line.

    Examples:
        - Some isolated hints (`hint_3` and `hint_5` are centrifugated).
        ```python
        foo = bar # paroxython: hint_1 hint_2
        # paroxython: hint_3
        fizz = [
            "foo",
            "bar", # paroxython: hint_4
            # paroxython: hint_5
        ]
        ```
        ```python
        foo = bar # paroxython: hint_1 hint_2 hint_3... hint_5...
        fizz = [
            "foo",
            "bar", # paroxython: hint_4
        ] # paroxython: ...hint_3 ...hint_5
        ```
        - Some trailing isolated hints (`hint_5` and `hint_6` are centrifugated).
        ```python
        foo = bar # paroxython: hint_1 hint_2
        fizz = [
            "foo",
            "bar", # paroxython: hint_3
        ] # paroxython: hint_4
        # paroxython: hint_5
        # paroxython: hint_6
        ```
        ```python
        foo = bar # paroxython: hint_1 hint_2 hint_5... hint_6...
        fizz = [
            "foo",
            "bar", # paroxython: hint_3
        ] # paroxython: hint_4 ...hint_5 ...hint_6
        ```
        - No isolated hints (no centrifugation of the existing hints).
        ```python
        foo = bar # paroxython: hint_1 hint_2
        fizz = [
            "foo",
            "bar", # paroxython: hint_4
        ]
        ```
        ```python
        foo = bar # paroxython: hint_1 hint_2
        fizz = [
            "foo",
            "bar", # paroxython: hint_4
        ]
        ```

        All examples above automatically extracted from
        [test_centrifugate_hints.py](https://repo/tests/test_centrifugate_hints.py).

        Argument `match_isolated_hints` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)

    <center>
    <iframe src='https://gfycat.com/ifr/ExcellentKeyElver' frameborder='0' scrolling='no' width='320' height='266'></iframe>
    </center>
    """
    lines = []
    hints: Set[str] = set()
    for line in source.split("\n"):
        m = match_isolated_hints(line)
        if m:
            hints.update(m[1].split())
        else:
            lines.append(line)
    if not hints:
        return source
    for i in (0, -1):
        if f" {HINT_COMMENT}" not in lines[i]:
            lines[i] += f" {HINT_COMMENT}"
    for hint in sorted(hints):
        lines[0] += f" {hint}..."
        lines[-1] += f" ...{hint}"
    return Source("\n".join(lines))


class HintBuffer:
    """Layer on the operations of appending, opening and closing a hint, used by `collect_hints`."""

    def __init__(self, description: str) -> None:
        self.result: LabelsSpans = defaultdict(list)
        self.stack: Dict[LabelName, List[int]] = defaultdict(list)
        self.description = description

    def append_hint(self, label_name: LabelName, line_number: int) -> None:
        """Label a span consisting in the given line number."""
        self.result[label_name].append(Span(line_number, line_number))

    def open_hint(self, label_name: LabelName, line_number: int) -> None:
        """Store the line number associated with a given opening hint."""
        self.stack[label_name].append(line_number)

    def get_champion(self, label_name: LabelName) -> List[Tuple[int, "HintBuffer"]]:
        """Return the last line number (if any) of the opening label, plus itself."""
        return [(self.stack[label_name][-1], self)] if self.stack[label_name] else []

    def close_hint(self, label_name: LabelName, line_number: int) -> None:
        """Label a span opened above and closed on the given line number."""
        self.result[label_name].append(Span(self.stack[label_name].pop(), line_number))

    def ensure_stack_is_empty(self) -> None:
        """Raise an error iff there remains at least one open label."""
        if any(self.stack.values()):
            print_fail(f"Unmatched opening hints for {self.description}: {self.stack}.")

    def get_result(self) -> LabelsSpans:
        """Return the result with sorted spans."""
        return {k: sorted(v) for (k, v) in self.result.items()}


def collect_hints(
    source: Source,
    match_label: Callable = regex.compile(r"^((?:-|\+|\.\.\.|…)?)(\w.*?)((?:\.\.\.|…)?)$").match,
) -> Tuple[LabelsSpans, LabelsSpans]:
    """Schedule for addition or deletion the hints appearing in the comments.

    Description:
        On this stage, the source code is scanned for manual hints. This function set them aside in
        two separate containers (depending on whether the user intends to add or remove them). They
        will later be used to tweak the labelling results obtained by static analysis.

    Args:
        source (Source): A source code.
        match_label (Callable, optional): A function matching a string:

            - optionally starting with `"+"`, `"-"`, `"..."` or `"…"`;
            - continuing with one alphanumeric character (`"[_0-9a-zA-Z]"`) and optionally a
              sequence of characters other than a newline (together they constitute the label);
            - optionally ending with `"..."` or `"…"`.

            [Not to be explicitly provided.](developer_manual/index.html#default-argument-trick)

    Raises:
        ValueError: Raised in various cases of malformed input.

    Returns:
        Tuple[LabelsSpans, LabelsSpans]: A couple of dictionaries associating labels with lists of
            spans. The first [resp. second] member collects the labels that the source code author
            has manually added [resp. suppressed].

    Example:
        >>> source = '''
        ...     1   foo() # paroxython: +hint_1...
        ...     2   bar() # paroxython: hint_2 hint_3 -hint_4
        ...     3   biz() # paroxython: ...hint_1
        ...     4   buz() # paroxython: -hint_2
        ... '''
        >>> collect_hints(source)
        (
            {
                "hint_1": [Span(start=1, end=3)],
                "hint_2": [Span(start=2, end=2)],
                "hint_3": [Span(start=2, end=2)],
            },
            {
                "hint_2": [Span(start=4, end=4)],
                "hint_4": [Span(start=2, end=2)],
            },
        )
        ```
    """
    addition = HintBuffer("addition")
    deletion = HintBuffer("deletion")
    for (i, line) in enumerate(source.split("\n"), 1):
        (_, separator, hints) = line.partition(f"{HINT_COMMENT} ")
        if not separator:
            continue
        for hint in hints.split():
            m = match_label(hint)
            if m is None:
                print_fail(f"Malformed hint '{hint}' on line {i}.")
            (before, label, after) = m.groups()
            if before in ("", "+", "-"):
                buffer = deletion if before == "-" else addition
                if after:
                    buffer.open_hint(label, i)
                else:
                    buffer.append_hint(label, i)
            else:  # `before` is in ("...", "…")
                if after:
                    print_fail(f"Illegal last part for hint {m.groups()} on line {i}.")
                champions = addition.get_champion(label) + deletion.get_champion(label)
                if not champions:
                    print_fail(f"Unmatched closing hint {m.groups()} on line {i}.")
                max(champions)[1].close_hint(label, i)
    addition.ensure_stack_is_empty()
    deletion.ensure_stack_is_empty()
    return (addition.get_result(), deletion.get_result())


def remove_hints(
    source: Source,
    sub_hints: Callable = regex.compile(fr"\s*{HINT_COMMENT} .*").sub,
) -> Source:
    """Once they are collected, remove all Paroxython hints from the given source.

    Argument `sub_hints` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
    """
    return Source(sub_hints("", source).strip())



"""
# Will you ever read me?

.. include:: ../README.md
   :start-after: </p>
   :end-before: # Read them

<br>

.. include:: ../docs/md/about.md

*[AST]: Abstract Syntax Tree
*[SQL]: Structured Query Language
*[TSV]: Tab Separated Values
*[JSON]: JavaScript Object Notation
*[SLOC]: Source Lines Of Code
"""

import sys
import pkg_resources

try:
    PAROXYTHON_VERSION = pkg_resources.get_distribution("paroxython").version
except pkg_resources.DistributionNotFound:
    PAROXYTHON_VERSION = "(unknown version)"  # For tests during CI

if "ipykernel" in sys.modules:
    # Declare the magic stuff paroxython iff the module is loaded as an IPython extension.

    from IPython.core.magic import Magics, cell_magic, magics_class  # type: ignore
    from IPython.display import Markdown, display  # type: ignore

    from .cli_tag import main

    def load_ipython_extension(ipython):
        ipython.register_magics(ParoxythonMagics)
        print(f"Paroxython {PAROXYTHON_VERSION} loaded.")

    @magics_class
    class ParoxythonMagics(Magics):
        @staticmethod
        @cell_magic
        def paroxython(line, cell=""):
            """
            Tag a Python code cell and output the table of its taxa or labels.

            In cell mode:
                %%paroxython [labels] [raw]
                ... Python cell code ...
            """
            args = set(line.lower().split())
            tags = "Label" if args.intersection(["label", "labels"]) else "Taxon"
            cell = f"# Compensate the offset of the magic command.\n{cell}"
            try:
                result = main(cell, tags=tags)
            except ValueError as e:
                print(e, file=sys.stderr)
            else:
                if "raw" in args:
                    print(result)
                else:
                    display(Markdown(result))



"""Scan a directory for Python programs and yield the corresponding `Program` objects."""

from pathlib import Path
from typing import Iterator

import regex  # type: ignore

from .preprocess_source import Cleanup, centrifugate_hints, collect_hints, remove_hints
from .user_types import Program, ProgramPath, Programs, Source


def list_programs(
    directory: Path,
    cleanup_strategy: str = "full",
    glob_pattern: str = "",
    skip_pattern: str = "",
    **kwargs,
) -> Programs:
    """List (by default recursively) all Python programs of a given directory.

    Args:
        directory (Path): The directory to search.
        cleanup_strategy (str, optional): Describes how to clean the source codes. Passed to
            `paroxython.preprocess_source.Cleanup`. Defaults to `"full"`.
        glob_pattern (str, optional): Describes which files to yield in `directory`. Passed to the
            standard library `pathlib`'s
            [`Path.glob()`](https://docs.python.org/3/library/pathlib.html#pathlib.Path.glob).
            If empty, replaced by `"**/*.py"`, which means “all Python source files in `directory`
            and all its subdirectories, recursively”. Defaults to `""`.
        skip_pattern (str, optional): Describes how to filter out the yielded files. If empty,
            skip any file whose name is `"__init__.py"`, `"setup.py"` or ends with `"-test.py"`,
            `"-tests.py"`, `"_test.py"` or `"_tests.py"`. Defaults to `""`.
        **kwargs: Ignored keyword arguments.

    Returns:
        Programs: A list of `Program` objects as constructed by `get_program`.
    """
    result: Programs = []
    cleanup = Cleanup(cleanup_strategy)
    glob_pattern = glob_pattern or "**/*.py"
    skip_pattern = skip_pattern or r"(__init__|setup|.*[-_]tests?)\.py"
    match_excluded = regex.compile(skip_pattern).fullmatch
    for path in sorted(directory.glob(glob_pattern)):
        if not match_excluded(path.name):
            source = cleanup.run(Source(path.read_text()))
            relative_path = path.relative_to(directory)
            result.append(get_program(source, relative_path))
    return result


def get_program(source: Source, relative_path: Path = None) -> Program:
    """Construct a fresh `Program` object from its source code and (optionally) relative path.

    Description:
        At this stage, the source code is already preprocessed by `paroxython.preprocess_source`,
        which among other things means it has been stripped from all its comments, except those
        consisting of one or several manual hints. The following operations are then carried out:

        1. Centrifugate the all-encompassing hints found in the source code (see
            `paroxython.preprocess_source.centrifugate_hints`).
        2. Collect all hints, determining whether they must be added to or removed from the
            labels which will later be found by `paroxython.label_programs.labelled_programs`.
        3. Remove all hints from the source code.
        4. Return a new `Program` (details below).

    Args:
        source (Source): A source code, already preprocessed.
        relative_path (Path, optional): A path relative to the directory passed to
            `list_programs()`. Can be omitted for testing purposes or when the source code is
            the contents of a Jupyter Notebook's cell. Defaults to None.

    Returns:
        Program: A `NamedTuple` consisting in the following fields:
        - `path` (type `ProgramPath`, derived from `str`): the program path, either empty or
            relative to the directory passed to `list_programs()`;
        - `source` (type `Source`, derived from `str`): its code source, fully cleaned up;
        - `addition` (type `Dict[LabelName, List[Span]]`): the manual hints scheduled for addition;
        - `deletion` (type `Dict[LabelName, List[Span]]`): the manual hints scheduled for deletion;
        - `labels` (type `List[Label]`, where each `Label` consists in a label name and a list of
            spans). This list is created empty here, to be later populated by
            `paroxython.label_programs.labelled_programs`.
        - `taxa` (type `List[Taxa]`, where each `Taxon` consists in a taxon name and a bag of
            spans). This list is created empty here, to be later calculated from the labels by
            `paroxython.map_taxonomy.Taxonomy`.
    """
    source = centrifugate_hints(source)
    (addition, deletion) = collect_hints(source)
    source = remove_hints(source)
    return Program(
        path=ProgramPath(str(relative_path or Path())),
        source=source,
        addition=addition,
        deletion=deletion,
        labels=[],
        taxa=[],
    )


def iterate_and_print_programs(programs: Programs) -> Iterator[Program]:
    """Iterate on a list of programs while printing their names as a side effect.

    Args:
        programs (Programs): A list of `Program` objects, as returned by `list_programs`.

    Yields:
        Iterator[Program]: The same `Program` objects in the same order.

    Note:
        This simple wrapper should print the name of each program over the previous one to avoid
        flooding the screen. Unfortunately, this may result in even more flood if your console
        does not support [ANSI escape codes](https://en.wikipedia.org/wiki/ANSI_escape_code).
    """
    blanks = ""
    for (i, program) in enumerate(programs, 1):
        print(end=f"\r{blanks}\r{i: 5} {program.path}", flush=True)
        blanks = " " * (len(program.path) + 7)
        if i == len(programs):  # Placed after the loop, the next line would not be executed.
            print(end=f"\r{blanks}\r", flush=True)
        yield program


if __name__ == "__main__":
    datetime = __import__("datetime").datetime
    directory = Path("../Algo/programs/")
    for program in list_programs(
        directory,
        skip_pattern=r"^(__init__|setup|.*[-_]tests?|quiz_.*)\.py$",
        cleanup_strategy="full",
    ):
        path = directory / program.path
        print(datetime.fromtimestamp(path.stat().st_mtime))
        print(program.source)
        print("-" * 80)



"""Collect all infos pertaining to the programs, labels and taxa into a so-called tag database."""

import json
import sqlite3
from bisect import insort
from collections import defaultdict
from datetime import datetime
from functools import lru_cache
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Dict

import regex  # type: ignore

from .goodies import add_line_numbers, print_success
from .label_programs import labelled_programs, iterate_and_print_programs
from .map_taxonomy import Taxonomy
from .user_types import (
    LabelInfos,
    Labels,
    LabelsPoorSpans,
    ProgramInfos,
    ProgramPath,
    ProgramPathSet,
    Programs,
    ProgramToPrograms,
    TaxonInfos,
    Taxa,
    TaxaPoorSpans,
)

__pdoc__ = {
    "TagDatabase": "",
    "TagDatabase.__init__": True,
}


class TagDatabase:
    def __init__(self, directory: Path, ignore_timestamps: bool = False, **kwargs) -> None:
        """Construct and populate the complete database of program features.

        Args:
            directory (Path): The directory to walk (by default, recursively), containing the
                Python programs of interest.
            ignore_timestamps (bool, optional): Don't store the last modification dates of these
                programms. Since the database snapshots generated by our tests are placed under
                version control, this prevents Git from tracking irrelevant differences. Defaults
                to `False`.
            **kwargs: May include the keyword arguments:

                - `print_performances`, transmitted to
                  `paroxython.label_programs.labelled_programs`;
                - `cleanup_strategy`, `skip_pattern`, `glob_pattern`, transmitted to
                  `paroxython.list_programs.list_programs` through
                  `paroxython.label_programs.labelled_programs`;
                - `taxonomy_path`, transmitted to `paroxython.map_taxonomy.Taxonomy`.

        Description:
            1. Create a list `programs` of `Program` objects, labelled by
              `paroxython.label_programs.labelled_programs`.
            2. Create a dictionary `self.labels` mapping each label name to the names of the
              programs featuring it (`collect_labels`).
            3. Map the labels to taxa (`paroxython.map_taxonomy.Taxonomy`), and populate the field
              `taxa` of `programs` (`map_labels_on_taxa`).
            4. Create a dictionary `self.taxa` mapping each taxon name to the names of the
              programs featuring it (`collect_taxa`).
            5. Create a dictionary `self.importations` mapping each program path to the paths
              of the programs importing it, both directly (`compute_direct_importations`) and
              indirectly (`complete_and_collect_importations`).
            6. Invert the previous dictionary to create a dictionary `self.exportations`
              (`compute_and_collect_exportations`).
            7. Gather in `self.programs_infos` each program timestamp, source, labels and taxa as
              a serialization-ready dictionary indexed by program paths.
        """

        self.directory = directory
        programs: Programs = labelled_programs(directory, **kwargs)
        self.labels = collect_labels(programs)

        map_labels_on_taxa(programs, Taxonomy(**kwargs))
        self.taxa = collect_taxa(programs)

        importations = compute_direct_importations(programs)
        self.importations = complete_and_collect_importations(importations)
        self.exportations = compute_and_collect_exportations(programs, self.importations)

        get_timestamp = lambda path: str(datetime.fromtimestamp(path.stat().st_mtime))
        if ignore_timestamps:
            get_timestamp = lambda path: ""
        self.programs_infos: ProgramInfos = {}
        for program in programs:
            self.programs_infos[program.path] = {
                "timestamp": get_timestamp(directory / program.path),
                "source": program.source,
                "labels": prepared_labels(program.labels),
                "taxa": prepared_taxa(program.taxa),
            }

    def get_json(self) -> str:
        r"""Dump the constructed `TagDatabase` object as a JSON string.

        Description:
            Schema and purpose [in the user manual](user_manual/index.html#the-json-database).

        Example:
            See the [JSON tag database](https://repo/examples/mini/programs_db.json) constructed from
            the programs of [this](https://repo/examples/mini/programs) directory.

        Note:
            For readability purposes, the output of `json.dumps()` is reformatted to fit each span
            list on a single line, e.g.:
            ```json
                        "flow/loop/exit/late": [
                            [
                                3,
                                8
                            ],
                            [
                                6,
                                7
                            ]
                        ],
            ```
            ... is unwrap as:
            ```json
                        "flow/loop/exit/late": [[3,8],[6,7]],
            ```
        """
        data = {
            "programs": self.programs_infos,
            "labels": dict(sorted(self.labels.items())),
            "taxa": dict(sorted(self.taxa.items())),
            "importations": dict(self.importations.items()),
            "exportations": dict(self.exportations.items()),
        }
        text = json.dumps(data, indent=2) + "\n"
        text = regex.sub(r"\s*\[\s+(\d+),\s+(\d+)\s+\](,?)\s+", r"[\1,\2]\3", text)
        return text

    def write_json(self, db_path: Optional[Path] = None) -> None:
        """Call `TagDatabase.get_json` and write the result to a file.

        Args:
            db_path (Optional[Path], optional): If not provided, and the directory provided to the
                class constructor is `"foobar"`, falls back to `"foobar_db.json"` in the same parent
                directory. Defaults to `None`.
        """
        db_path = db_path or self.directory.parent / f"{self.directory.name}_db.json"
        db_path.write_text(self.get_json())
        print_success(f"Dumped: {db_path.absolute()}.")

    def write_sqlite(self, db_path: Optional[Path] = None) -> None:
        """Dump the constructed `TagDatabase` object as a SQLite database (experimental).

        Args:
            db_path (Optional[Path], optional): The path of the SQLite database. If not provided,
                and the directory provided to the class constructor is `"foobar"`, falls back to
                `"foobar_db.sqlite"` in the same parent directory. Defaults to `None`.

        Description:
            Schema, purpose and example queries
            [in the user manual](user_manual/index.html#the-sqlite-database).
        """
        db_path = db_path or self.directory.parent / f"{self.directory.name}_db.sqlite"
        program_rows: List[Tuple] = []
        label_rows: List[Tuple] = []
        taxon_rows: List[Tuple] = []
        for (path, info) in self.programs_infos.items():
            source = f"{path}\n\n" + add_line_numbers(info["source"])
            program_rows.append((path, info["timestamp"], source))
            for (label_name, spans) in info["labels"].items():
                (prefix, _, suffix) = label_name.partition(":")
                for span in spans:
                    label_rows.append(
                        (
                            label_name,
                            prefix,
                            suffix,
                            "-".join(map(str, span)) if span[0] != span[1] else str(span[0]),
                            span[0],
                            span[1],
                            path,
                        )
                    )
            for (taxon_name, spans) in info["taxa"].items():
                for span in spans:
                    taxon_rows.append(
                        (
                            taxon_name,
                            "-".join(map(str, span)) if span[0] != span[1] else str(span[0]),
                            span[0],
                            span[1],
                            path,
                        )
                    )

        if db_path.exists():  # pragma: no cover / Python 3.8: use missing_ok=True parameter
            db_path.unlink()
        connexion = sqlite3.connect(str(db_path))  # str() for Python 3.6 compatibility
        c = connexion.cursor()

        fill = lambda columns: ",".join("?" * len(columns))
        program_columns = (
            "program TEXT PRIMARY KEY",
            "timestamp TEXT",
            "source TEXT",
        )
        c.execute(f"CREATE TABLE program ({','.join(program_columns)})")
        c.executemany(f"INSERT INTO program VALUES ({fill(program_columns)})", program_rows)

        label_columns = (
            # use rowid as primary key
            "label TEXT",
            "label_prefix TEXT",
            "label_suffix TEXT",
            "span TEXT",
            "span_start INTEGER",
            "span_end INTEGER",
            "program TEXT",
        )
        c.execute(
            f"CREATE TABLE label ({','.join(label_columns)},"
            "FOREIGN KEY (program) REFERENCES program (program))"
        )
        c.executemany(f"INSERT INTO label VALUES ({fill(label_columns)})", label_rows)

        taxon_columns = (
            # use rowid as primary key
            "taxon TEXT",
            "span TEXT",
            "span_start INTEGER",
            "span_end INTEGER",
            "program TEXT",
        )
        c.execute(
            f"CREATE TABLE taxon ({','.join(taxon_columns)},"
            "FOREIGN KEY (program) REFERENCES program (program))"
        )
        c.executemany(f"INSERT INTO taxon VALUES ({fill(taxon_columns)})", taxon_rows)

        connexion.commit()
        connexion.close()

        print_success(f"Dumped: {db_path.absolute()}.")


def collect_labels(programs: Programs) -> LabelInfos:
    """Iterate through the given programs to collect all their labels.

    Args:
        programs (Programs): A list of `Program` objects, with their `name` and `labels` fields
            already populated.

    Returns:
        LabelInfos: A dictionary mapping each label name to the names of the programs featuring it.
    """
    result: LabelInfos = defaultdict(list)
    for program in programs:
        for label in program.labels:
            result[label.name].append(program.path)
    return result


def map_labels_on_taxa(programs: Programs, taxonomy: Taxonomy) -> None:
    """Populate the `taxa` fields of the given `programs` with the translations of their labels.

    Args:
        programs (Programs): A list of `Program` objects with their `labels` field populated.
        taxonomy (Taxonomy): A “translator” of label names into taxon names.
    """
    print(f"Mapping taxonomy on {len(programs)} programs.")
    for program in iterate_and_print_programs(programs):
        program.taxa[:] = taxonomy.to_taxa(program.labels)
        # `program` being a tuple, modifying its fields can only be done in place.


def collect_taxa(programs: Programs) -> TaxonInfos:
    """Iterate through the given programs to collect all their taxa.

    Args:
        programs (Programs): A list of `Program` objects, with their `name` and `taxa` fields
            already populated.

    Returns:
        TaxonInfos: A dictionary mapping each taxon name to the names of the programs featuring it.
    """
    result: TaxonInfos = defaultdict(list)
    for program in programs:
        for taxon in program.taxa:
            result[taxon.name].append(program.path)
    return result


def compute_direct_importations(
    programs: Programs, match_import: Callable = regex.compile(r"import_internally:([^:]+)").match
) -> ProgramToPrograms:
    """Associate each given labelled program to the set of its direct internal imports.

    Description:
        Iterate through the programs, retrieve their labels starting with `"import_internally"`
        and collect the names of the involved programs.

    Args:
        programs (Programs): A list of labelled `Program` objects.
        match_import (Callable, optional):  A function taking a label name and, in
            the case it starts with `"import_internally:"`, returns a match object whose
            first group is the name of the imported program.
            [Not to be explicitly provided.](developer_manual/index.html#default-argument-trick)

    Returns:
        ProgramToPrograms: A dictionary mapping every program path to the list of the paths of
            the internal programs it imports directly.
    """
    importations: Dict = {program.path: set() for program in programs}
    for program in programs:
        for label in program.labels:
            match = match_import(label.name)
            if match:  # Python 3.8: use assignement-expression
                importations[program.path].add(f"{match[1]}.py")
    return importations


def complete_and_collect_importations(importations: ProgramToPrograms) -> ProgramToPrograms:
    """Complete the direct internal imports with indirect ones.

    Args:
        importations (ProgramToPrograms): A dictionary mapping every program path to the list
            of the paths of the internal programs it imports directly.

    Returns:
        ProgramToPrograms: A dictionary mapping every program path to the sorted list of the
            paths of the internal programs it imports either directly or indirectly.
    """

    @lru_cache(maxsize=None)
    def complete_internal_imports(program_path: ProgramPath) -> ProgramPathSet:
        result: ProgramPathSet = set(importations.get(program_path, []))
        for imported in list(result):  # traverse a copy
            result.update(complete_internal_imports(imported))
        return result

    completed_importations: ProgramToPrograms = {}
    for program_path in list(importations.keys()):
        completed_importations[program_path] = sorted(complete_internal_imports(program_path))
    return completed_importations


def compute_and_collect_exportations(
    programs: Programs, importations: ProgramToPrograms
) -> ProgramToPrograms:
    """Invert `importations` to construct `exportations`.

    Args:
        programs (Programs):  A list of named `Program` objects.
        importations (ProgramToPrograms): A dictionary mapping every program path to the list of
            the paths of the internal programs it imports either directly or indirectly.

    Returns:
        ProgramToPrograms: A dictionary mapping every program path to the sorted list of the paths
            of the internal programs it is imported by, either directly or indirectly.
    """
    exportations: ProgramToPrograms = {program.path: [] for program in programs}
    for (importing_path, imported_paths) in importations.items():
        for imported_path in imported_paths:
            if importing_path not in exportations[imported_path]:
                insort(exportations[imported_path], importing_path)
    return exportations


def prepared_labels(labels: Labels) -> LabelsPoorSpans:
    """Prepare the labels for serialization.

    Args:
        labels (Labels): The list of labels to be serialized. Each label consists in a name and a
            **list** of spans. Each span is a tuple starting by the actual range of line numbers.

    Returns:
        LabelPoorSpans:
            A dictionary mapping the label names with the list of their spans, transformed into
            simple couples of integers.

    Example:
        >>> prepared_labels([
        ...     ("name_1", [
        ...             Span(start=1, end=2, path="foo"),
        ...             Span(start=1, end=2, path="foo"),  # note the duplicate
        ...             Span(start=1, end=2, path="bar"),  # note the difference
        ...         ]),
        ...     ("name_2", [
        ...             Span(start=2, end=4, path="fizz"),
        ...             Span(start=6, end=7, path="buzz"),
        ...         ]),
        ...     ("name_3", [Span(start=5, end=5, path="foobar")]),
        ... ])
        {
            "name_1": [(1, 2), (1, 2)],  # deduplicated
            "name_2": [(2, 4), (6, 7)],
            "name_3": [(5, 5)],
        }
    """
    result: LabelsPoorSpans = {}
    for (label_name, spans) in labels:
        result[label_name] = [(span.start, span.end) for span in sorted(set(spans))]
    return result


def prepared_taxa(taxa: Taxa) -> TaxaPoorSpans:
    """Prepare the taxa for serialization.

    Args:
        taxa (Taxa): The list of taxa to be serialized. Each taxon consists in a name and a
            **bag** of spans. Each span is a tuple starting by the actual range of line numbers.

    Returns:
        TaxonPoorSpans:
            A dictionary mapping the taxon names with the list of their spans, transformed into
            simple couples of integers.

    Example:
        >>> prepared_taxa([
        ...     ("name_1", Counter({
        ...             Span(start=1, end=2, path="foo"): 2,  # note the duplicate
        ...             Span(start=1, end=2, path="bar"): 1,
        ...         })),
        ...     ("name_2", Counter({
        ...             Span(start=2, end=4, path="fizz"): 1,
        ...             Span(start=6, end=7, path="buzz"): 1,
        ...         })),
        ...     ("name_3", Counter({Span(start=5, end=5, path="foobar"): 1}))
        ... ])
        {
            "name_1": [(1, 2), (1, 2)],  # deduplicated
            "name_2": [(2, 4), (6, 7)],
            "name_3": [(5, 5)],
        }
    """
    result: TaxaPoorSpans = {}
    for (taxon_name, spans) in taxa:
        result[taxon_name] = [(span.start, span.end) for span in sorted(spans)]
    return result


if __name__ == "__main__":
    # fmt:off
    directories = [
        # "examples/simple/programs",
        # "../algo/programs",
        "../Python",
    ]
    # fmt:on
    print()
    for directory in directories:
        db = TagDatabase(Path(directory), ignore_timestamps=True)
        db.write_json()
        db.write_sqlite()
    print()



"""Tag one Python file and output the table of its taxa or labels.

USAGE:
    ```
    paroxython tag [options] FILENAME
    ```

OPTIONS:
    ```plain
    -f --format=FORMAT  Format of the output, either "md" (Markdown)
                        or "tsv" (Tab Separated Values). [default: md]
    -l --labels         Output the labels instead of the taxa.
    -t --taxonomy=PATH  The path of a TSV file mapping labels onto taxa.
                        If not specified, use the included default taxonomy:
                        https://github.com/laowantong/paroxython/blob/0.6.1/paroxython/resources/taxonomy.tsv
    ```
"""

from pathlib import Path
from typing import List, Tuple, Optional
from typing_extensions import Literal

from .goodies import couple_to_string, print_exit
from .parse_program import ProgramParser
from .list_programs import get_program
from .map_taxonomy import Taxonomy
from .user_types import Source


def main(
    source: Source,
    tags: Literal["Taxon", "Label"] = "Taxon",
    relative_path: Path = Path("."),
    output_format: Literal["md", "tsv"] = "md",
    taxonomy_path: Optional[Path] = None,
) -> str:
    program = get_program(source, relative_path)
    parse = ProgramParser()
    program.labels[:] = parse(program)
    couples: List[Tuple[str, str]] = [(tags, "Lines")]
    if tags == "Label":
        for (label_name, label_spans) in sorted(program.labels):
            s = ", ".join(map(couple_to_string, sorted(label_spans)))
            couples.append((label_name, s))
    else:
        taxonomy = Taxonomy(taxonomy_path)
        taxa = taxonomy.to_taxa(program.labels)
        for (taxon_name, taxon_spans) in sorted(taxa):
            s = ", ".join(map(couple_to_string, sorted(taxon_spans.elements())))
            couples.append((taxon_name, s))
    if output_format == "md":
        result = [f"| `{tag}` | {lines} |" for (tag, lines) in couples]
        result[0] = result[0].replace("`", "")
        result[1:1] = ["|:--|:--|"]
    else:
        result = [f"{tag}\t{lines}" for (tag, lines) in couples]
    return "\n".join(result)


def cli_wrapper(args):
    path = Path(args["FILENAME"])
    try:
        source = path.read_text()
    except Exception as e:
        print_exit(str(e))
    result = main(
        source=Source(source),
        tags="Label" if args["--labels"] else "Taxon",
        relative_path=path.parent,
        output_format=args["--format"],
        taxonomy_path=Path(args["--taxonomy"]) if args["--taxonomy"] else None,
    )
    print(result)



"""List all the programs of a given directory, with their field `labels` populated."""

from pathlib import Path
from typing import Callable, Iterator, List, Set

import regex  # type: ignore

from .list_programs import iterate_and_print_programs, list_programs
from .parse_program import ProgramParser
from .user_types import Label, LabelName, Programs, Source


def labelled_programs(
    directory: Path,
    search_imported_program_path: Callable = regex.compile(r"import(?:_module)?:([^:]+)").search,
    print_performances: bool = False,
    **kwargs,
) -> Programs:
    """Walk a given directory, label all its programs, and return a list of them.

    Args:
        directory (Path): The directory to walk, containing some Python programs.
        search_imported_program_path (Callable, optional): A function taking a label name and, in
            the case it starts with `"import:"` or `"import_module:"`, returns a match object
            whose first group is the name of the imported program.
            [Not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
        print_performances (bool): If `True`, ends with a call to
            `paroxython.parse_program.ProgramParser.print_performances`. Defaults to `False`.
        **kwargs: May include the keyword arguments `cleanup_strategy`, `skip_pattern`,
            `glob_pattern`, transmitted to `paroxython.list_programs.list_programs`.
    Note:
        In addition to creating a list of programs with a populated `labels` field, tweak all of
        those which mark an importation of an “internal” module, _i.e_, a program belonging to
        this very list. For instance, a label `"import:my_program"` would be transformed into
        `"import_internally:my_program"`, while `"import:itertools"` would be left untouched.
    """
    programs: Programs = list_programs(directory, **kwargs)
    internal_program_paths = {p.path.replace("/", ".") for p in programs}  # path sep -> import sep
    parse = ProgramParser()
    print(f"Labelling {len(programs)} programs.")
    for program in iterate_and_print_programs(programs):
        program.labels[:] = parse(program)  # populate this field in place with [:]
        for (i, label) in enumerate(program.labels):
            m = search_imported_program_path(label.name)
            if m and f"{m[1]}.py" in internal_program_paths:
                tweaked_label_name = label.name.replace(":", "_internally:", 1)
                tweaked_label_name = tweaked_label_name.replace(".", "/")  # import sep -> path sep
                program.labels[i] = Label(name=LabelName(tweaked_label_name), spans=label.spans)
    if print_performances:  # pragma: no cover
        parse.print_performances()
    return programs


def generate_labelled_sources(programs: Programs) -> Iterator:
    """For each program, yield its source with its labels in comment.

    Args:
        programs (Programs): A list of labelled programs.

    Yields:
        Iterator: The lines of the given programs, with their labels in comment.

    Note:
        This function is for testing purposes only. See an example of the result in
        [labelled_sources.py](https://repo/examples/mini/labelled_sources.py).
    """
    separator = "-" * 88
    for program in programs:
        yield Source(f"# {separator}\n# {program.path}\n# {separator}")
        lines = program.source.splitlines()
        if lines:
            comments: List[Set[str]] = [set() for _ in lines]
            for label in program.labels:
                for span in label.spans:
                    span_length = span.end - span.start
                    span_suffix = f" (-> +{span_length})" if span_length else ""
                    comments[span.start - 1].add(f"{label.name}{span_suffix}")
            for (i, comment) in enumerate(comments):
                if comment:
                    lines[i] += " # " + ", ".join(sorted(comment))
        yield Source("\n".join(lines + [""]))


if __name__ == "__main__":
    programs = labelled_programs(Path("docs/resources"))
    for result in generate_labelled_sources(programs):
        print(result)



"""Evolve a set of selected programs and a set of taxa representing the imparted knowledge.

The class stores the state of these two sets and provides a collection of operators to be applied
on them during the execution of the pipeline (see `paroxython.recommend_programs`). Two additional
sets, less important, are maintained: the programs and the taxa to be hidden during the display of
the final result.

.. image:: resources/filter_flow.png

.. note::
    Initially, all programs in the tag database are selected. The three others sets are empty. As
    the filter evolves, the first set can only shrink, and the other ones only increase.
"""

from collections import Counter as counter
from collections import defaultdict
from itertools import permutations, product
from typing import Counter, Dict, Iterator, List, Tuple

import regex  # type: ignore

from .goodies import print_warning
from .normalize_predicate import normalize_predicate
from .user_types import (
    Criterion,
    JsonTagDatabase,
    Operation,
    Predicate,
    ProgramInfos,
    ProgramPath,
    ProgramPathSet,
    ProgramToPrograms,
    TaxonInfos,
    TaxonName,
    TaxonNameSet,
    TaxaPoorSpans,
)

__pdoc__ = {
    "ProgramFilter.__init__": True,
    "ProgramFilter": "",
}


class ProgramFilter:

    # Initialization of the filter

    def __init__(self, db: JsonTagDatabase) -> None:
        """Call `define_shortcuts()`,  `init_filter_state()` and `add_imported_taxa()` (below)."""
        self.define_shortcuts(db)
        self.init_filter_state()
        self.add_imported_taxa()

    def define_shortcuts(self, db: JsonTagDatabase) -> None:
        """Define some attributes that point directly to the main parts of the tag database."""
        self.db_programs: ProgramInfos = db["programs"]
        self.db_taxa: TaxonInfos = db["taxa"]
        self.db_importations: ProgramToPrograms = db["importations"]
        self.db_exportations: ProgramToPrograms = db["exportations"]

    def init_filter_state(self) -> None:
        """Select all programs and define the imparted knowledge as empty."""
        self.selected_programs: ProgramPathSet = set(self.db_programs)
        self.imparted_knowledge: TaxonNameSet = set()
        self.hidden_taxa: TaxonNameSet = set()
        self.hidden_programs: ProgramPathSet = set()

    def add_imported_taxa(self) -> None:
        """Copy under each program the taxa it featured by importation.

        Description:
            Initially, the JSON tag database stores under each program all the taxa it features
            _directly_, for instance:
            ```
            {
                "programs": {
                    ...
                    "fizzbuzz.py": {
                        ...
                        "taxa": {
                            "call/subroutine/builtin/print": [[4,4],[6,6],[8,8],[10,10]],
                            "call/subroutine/builtin/range": [[2,2]],
                            "flow/conditional": [[3,10],[5,10],[7,10]],
                            ...
            ```

             For performance and readability reasons, it is useful to complete such a dictionary
             with the taxa which are featured _indirectly_, namely those featured by the imported
             programs. For instance, assuming that `fizzbuzz.py` imports `collatz.py`, we must
             copy the taxa featured by the latter, but not by the former program, namely:

            ```
                            "operator/arithmetic/multiplication": [],
                            "def/argument/arg": [],
                            "def/subroutine/procedure": [],
                            ...
            ```

        Notes:
            - The spans of the imported taxa are not copied, but replaced by an empty list (it is
            enough to know that a certain taxon is only featured by an imported program, not on
            which lines of which program).
            - All imported taxa starting with `"meta/"` are by convention excluded from the
            copy.
        """
        for (exporter, importers) in self.db_exportations.items():
            exported_taxa = list(self.db_programs[exporter]["taxa"])
            for exported_taxon in exported_taxa:
                if exported_taxon.startswith("meta/"):
                    continue
                for importer in importers:
                    importer_taxa = self.db_programs[importer]["taxa"]
                    if exported_taxon not in importer_taxa:
                        importer_taxa[exported_taxon] = []

    # Get the set of taxon names or program paths matching the given pattern.

    def taxa_of_pattern(self, pattern: str) -> TaxonNameSet:
        """Find all the existing taxa matching the given regular expression pattern.

        Note:
            All taxa are searched, not just the ones featured by a selected program.
        """
        match = regex.compile(fr"{pattern}\b").match
        result: TaxonNameSet = set(filter(match, self.db_taxa))
        if not result:
            print_warning(f"the pattern '{pattern}' doesn't match any existing taxon.")
        return result

    def programs_of_pattern(self, pattern: str) -> ProgramPathSet:
        """Find all the existing programs matching the given regular expression pattern.

        Note:
            All programs are searched, not just the selected ones.
        """
        match = regex.compile(fr"{pattern}").match
        result: ProgramPathSet = set(filter(match, self.db_programs))
        if not result:
            print_warning(f"the pattern '{pattern}' doesn't match any existing program.")
        return result

    # Select programs from the taxa they feature, and vice versa.

    def taxa_of_programs(self, programs: ProgramPathSet, follow: bool = False) -> TaxonNameSet:
        """Return the taxa featured (or optionally imported) by any given program.

        Args:
            programs (ProgramPathSet): Program paths. The non-existing ones are silently ignored.
            follow (bool, optional): If true, include the taxa featured by the imported programs.
                Defaults to `False`.

        Example:
            Let `p1`, `p2`, `p3` be three programs, with `p3` importing `p2`, and `p2` importing
                `p1`. Let `t` be a taxon featured by `p1` only. Then:
            >>> t in taxa_of_programs({p1})
            True
            >>> t in taxa_of_programs({p2})
            False
            >>> t in taxa_of_programs({p3}, follow=True)
            True
        """
        taxa: TaxonNameSet = set()
        for program in programs:
            if program in self.db_programs:
                for (taxon, spans) in self.db_programs[program]["taxa"].items():
                    if spans or follow:
                        # Either the taxon is featured directly or the imports must be followed.
                        taxa.add(taxon)
        return taxa

    def programs_of_taxa(self, taxa: TaxonNameSet, follow: bool = False) -> ProgramPathSet:
        """Return the programs featuring (or optionally importing) any given taxon.

        Args:
            taxa (TaxonNameSet): Taxon names. The non-existing ones are silently ignored.
            follow (bool, optional): If true, include the programs importing at least one program
                featuring at least one taxon. Defaults to `False`.

        Example:
            With `p1`, `p2`, `p3` and `t` as in the example of `ProgramFilter.taxa_of_programs`:
            >>> programs_of_taxa({t})
            {p1}
            >>> programs_of_taxa({t}, follow=True)
            {p1, p2, p3}
        """
        programs: ProgramPathSet = set()
        for taxon in taxa:
            programs.update(self.db_taxa.get(taxon, []))
        if follow:
            for program in list(programs):
                programs.update(self.db_exportations[program])
        return programs

    # Update the state of the filter by applying set operations with the given programs or taxa.

    def update_filter(
        self,
        criteria: List[Criterion],
        operation: Operation,
        quantifier: str,
    ) -> None:
        """Update the selected programs and/or impart the associated taxa and/or mark them as hidden.

        Description:
            - If the operation is `"impart"`:
                - calculate the appropriate sets of programs and taxa;
                - remove these programs from `self.selected_programs`;
                - add these taxa to `self.imparted_knowledge`.
            - If the operation is `"hide"`:
                - calculate the appropriate sets of programs and taxa;
                - add these taxa to `self.hidden_taxa`.
            - Otherwise (the operation is either `"include"` or `"exclude"`):
                - calculate the appropriate bag of programs: this bag counts, for each program,
                    the number of criteria they meet (maximum: size of `criteria`);
                - if `quantifier` is `"all"`, remove from this bag all programs which do not
                    meet at least one criterion;
                - include or exclude the resulting programs. Note that the `"exclude"` operation
                    extends to the programs which import the resulting ones: if the user wants to
                    exclude a program, she obviously expects that the programs which require it
                    are excluded too.

        Args:
            criteria (List[Criterion]): A list of criteria, i.e., a mix of regular expression
                patterns (strings) and/or predicates (triples).
            operation (Operation): Either `"impart"`, `"hide"`, `"include"` or `"exclude"`.
            quantifier (str): Either `"any"` or `"all"`.
        """
        if operation in ("impart", "hide"):
            patterns = [str(criterion) for criterion in criteria]
            if operation == "impart":
                (program_set, taxon_set) = self.programs_and_taxa_of_patterns(patterns)
                self.exclude_programs(program_set, follow=False)
                self.impart_taxa(taxon_set)
            else:
                (program_set, taxon_set) = self.programs_or_taxa_of_patterns(patterns)
                self.hidden_programs.update(program_set)
                self.hidden_taxa.update(taxon_set)
        else:
            program_bag = self.programs_of_criteria(criteria, follow=(operation == "exclude"))
            if quantifier == "all":
                program_bag -= counter({program: len(criteria) - 1 for program in program_bag})
            if operation == "include":
                self.include_programs(set(program_bag))
            else:  # necessarily "exclude"
                self.exclude_programs(set(program_bag), follow=True)

    def programs_and_taxa_of_patterns(
        self, patterns: List[str]
    ) -> Tuple[ProgramPathSet, TaxonNameSet]:
        """Calculate the sets of programs and taxa matching at least one of the patterns.

        Description:
            Each pattern is a string which is interpreted either as:

            - a program path pattern (ending with `".py"`). All programs matching it are accumulated
                in the result, _along with any taxon they feature, directly or by importation, this
                being the only difference with `ProgramFilter.programs_or_taxa_of_patterns`_;
            - or a taxon name pattern. All taxa matching it are accumulated in the result.

        Args:
            patterns (List[str]): A list of regular expression patterns (strings).

        Returns:
            Tuple[ProgramPathSet, TaxonNameSet]: The couple of accumulated programs and taxa.
        """
        resulting_taxa: TaxonNameSet = set()
        resulting_programs: ProgramPathSet = set()
        for pattern in patterns:
            if pattern.endswith(".py"):
                programs = self.programs_of_pattern(pattern)
                taxa = self.taxa_of_programs(programs, follow=False)
                resulting_programs.update(programs)
            else:
                taxa = self.taxa_of_pattern(pattern)
            resulting_taxa.update(taxa)
        return (resulting_programs, resulting_taxa)

    def programs_or_taxa_of_patterns(
        self, patterns: List[str]
    ) -> Tuple[ProgramPathSet, TaxonNameSet]:
        """See `ProgramFilter.programs_and_taxa_of_patterns`, minus the part in _italics_."""
        resulting_taxa: TaxonNameSet = set()
        resulting_programs: ProgramPathSet = set()
        for pattern in patterns:
            if pattern.endswith(".py"):
                resulting_programs.update(self.programs_of_pattern(pattern))
            else:
                resulting_taxa.update(self.taxa_of_pattern(pattern))
        return (resulting_programs, resulting_taxa)

    def programs_of_criteria(self, criteria: List[Criterion], follow: bool) -> Counter[ProgramPath]:
        """Calculate the set of programs that meet at least one of the criteria.

        Description:
            Each criterion may be either:

            - a string, which will be interpreted either as:
                - a program path pattern (ending with `".py"`). All programs matching it are
                    accumulated in the result;
                - or a taxon name pattern. All programs featuring at least one taxon matching it
                    are accumulated in the result. If the operation is `"exclude"`, this set is
                    extended to the programs which import (either directly or by transitivity) at
                    least one of its members;
            - a triple consisting in a “subject” pattern, a predicate (positive or negative)
                and an ”object” pattern. This predicate is normalized and, depending on its
                “sign”, evaluated on the patterns by either `ProgramFilter.programs_of_triple` or
                `ProgramFilter.programs_of_negated_triple`.

        Args:
            criteria (List[Criterion]): A list of criteria, i.e., a mix of regular expression
                patterns (strings) and/or predicates (triples).
            follow (bool): If `True`, extend the result with all the programs which import (either
                directly or by transitivity) at least one program meeting a criterion.

        Returns:
            Counter[ProgramPath]: A bag (multiset) counting, for each resulting program, the number
                of criteria it meets.
        """
        resulting_programs: Counter[ProgramPath] = counter()
        for criterion in criteria:
            if isinstance(criterion, str):  # the criterion is a pattern
                if criterion.endswith(".py"):  # the pattern is a program pattern
                    programs = self.programs_of_pattern(criterion)
                else:  # the pattern is a label pattern
                    taxa = self.taxa_of_pattern(criterion)
                    programs = self.programs_of_taxa(taxa, follow=follow)
                resulting_programs.update(programs)
            elif isinstance(criterion, (list, tuple)) and len(criterion) == 3:
                (pattern_1, raw_predicate, pattern_2) = criterion
                (predicate, negated) = normalize_predicate(raw_predicate)
                function = self.programs_of_negated_triple if negated else self.programs_of_triple
                resulting_programs.update(function(pattern_1, predicate, pattern_2))
            else:
                print_warning(f"criterion {repr(criterion)} cannot be included or excluded.")
        return resulting_programs

    def impart_taxa(self, taxa: TaxonNameSet) -> None:
        """Enrich the imparted knowledge with all the prefixes of the given taxa.

        Add to `self.imparted_knowledge` the given taxa, along with all their prefixes. For
        instance, imparting a taxon `a/b/c`, will impart taxa `a` and `a/b` too.

        Args:
            taxa (TaxonNameSet): The concepts to be imparted along with their “super” concepts.
        """
        for taxon in taxa:
            edges = taxon.split("/")
            for i in range(len(edges)):
                prefix = "/".join(edges[: i + 1])
                self.imparted_knowledge.add(TaxonName(prefix))

    def include_programs(self, programs: ProgramPathSet) -> None:
        """Deselect the programs not found among the given ones.

        Args:
            programs (ProgramPathSet): The set of programs to keep (provided they are already in
                `self.selected_programs`). All other programs of `self.selected_programs` will be
                filtered out.
        """
        self.selected_programs.intersection_update(programs)

    def exclude_programs(self, programs: ProgramPathSet, follow: bool) -> None:
        """Deselect the programs found among the given ones or (optionally) importing them.

        Args:
            programs (ProgramPathSet): The set of programs to exclude.
            follow (bool): If `True`, exclude also all programs which import (either directly or
                by transitivity) at least one member of the given `programs`.
        """
        self.selected_programs.difference_update(programs)
        if follow:
            for program in programs:
                self.selected_programs.difference_update(self.db_exportations.get(program, {}))

    # Deal with semantic triples of the form (taxon_pattern_1, taxon_pattern_2, predicate)

    def programs_of_triple(
        self,
        taxon_pattern_1: str,
        predicate: Predicate,
        taxon_pattern_2: str,
    ) -> ProgramPathSet:
        """Return the programs where two given taxa satisfy a given predicate.

        Args:
            taxon_pattern_1 (str): A regular expression pattern matching the “subject” of the
                semantic triple.
            predicate (Predicate): the predicate of the semantic triple, always expressed in
                positive form.
            taxon_pattern_2 (str): A regular expression pattern matching the ”object” of the
                semantic triple.

        Returns:
            ProgramPathSet: The programs featuring at least one span `s_1` of `taxon_1` and one
                span `s_2` of `taxon_2` for which `predicate(s_1, s_2)` is verified.
        """
        taxa_1 = self.taxa_of_pattern(taxon_pattern_1)
        taxa_2 = self.taxa_of_pattern(taxon_pattern_2)
        programs_1 = self.programs_of_taxa(taxa_1)
        programs_2 = self.programs_of_taxa(taxa_2)
        result: ProgramPathSet = set()
        for program in programs_1 & programs_2:  # for each program featuring both taxon sets
            spans = self.db_programs[program]["taxa"]
            for (span_1, span_2) in self.iterate_on_spans(spans, taxa_1, taxa_2):
                if predicate(span_1, span_2):
                    result.add(program)
                    break
        return result

    def programs_of_negated_triple(
        self,
        taxon_pattern_1: str,
        predicate: Predicate,
        taxon_pattern_2: str,
    ) -> ProgramPathSet:
        r"""Return the programs where the given predicate is not satisfied.

        Args:
            The same arguments as `ProgramFilter.programs_of_triple()`, including the fact that the
            predicate is expressed in **positive** form.

        Returns:
            ProgramPathSet: The set of programs which feature at least one taxon matching
                `taxon_pattern_1`, and such that, for any  span `s_1` of such a taxon, there exists
                no span `s_2` of a taxon matching `taxon_pattern_2` for which `predicate(s_1, s_2)`
                is verified.

        Details:
            See the [pipeline documentation](user_manual/index.html#general-semantics-of-the-negation).
        """
        taxa_1 = self.taxa_of_pattern(taxon_pattern_1)
        taxa_2 = self.taxa_of_pattern(taxon_pattern_2)
        programs_1 = self.programs_of_taxa(taxa_1)
        programs_2 = self.programs_of_taxa(taxa_2)
        result: ProgramPathSet = programs_1  # by default, keep all programs featuring taxon_1
        for program in programs_1 & programs_2:  # for each program featuring both taxon sets
            spans = self.db_programs[program]["taxa"]
            exists_span_2_satisfying_predicate: Dict[Tuple, bool] = defaultdict(bool)
            for (span_1, span_2) in self.iterate_on_spans(spans, taxa_1, taxa_2):
                exists_span_2_satisfying_predicate[tuple(span_1)] |= predicate(span_1, span_2)
            if all(exists_span_2_satisfying_predicate.values()):
                # for any span_1, there is at least one span_2 such that predicate(span_1, span_2)
                result.remove(program)
        return result

    def iterate_on_spans(
        self,  # declared as an instance method to prevent pdoc3 from misplacing its documentation.
        spans: TaxaPoorSpans,
        taxa_1: TaxonNameSet,
        taxa_2: TaxonNameSet,
    ) -> Iterator:
        """Generate all relevant couples of spans for the given “subject” and ”object” taxa.

        Description:
            This private function is invoked by `ProgramFilter.programs_of_triple` and
            `ProgramFilter.programs_of_negated_triple` to enumerate the couples of spans on which
            a condition of the form : “`subject_span` `predicate` `object_span`” will be evaluated.
            Initially, such a condition is expressed by the user as: “`subject_pattern` `predicate`
            `object_pattern`”. Both patterns may match several taxa and, in a given program, each
            taxon can occur on several different or even identical spans. Ultimately, the predicate
            expresses a relation between two spans.

        Args:
            spans (TaxaPoorSpans): The dictionary of the taxa featured by a certain program,
                each taxon being associated with a list of poor spans (“poor” meaning that they
                contain no other information than the numbers of the first and last line of a taxon
                occurrence).
            taxa_1 (TaxonNameSet): The various “subject” taxa of the semantic triple.
            taxa_2 (TaxonNameSet): The various “object” taxa of the semantic triple.

        Yields:
            Iterator[PoorSpan]: Couples of poor spans, in no particular order.

        Example:
            Consider a program featuring taxa `t1`, `t2`, `t3` and `t4` on the following spans:
            >>> spans = {
            ...      "t1": [(1, 1), (1, 1), (1, 2)],
            ...      "t2": [(2, 3), (1, 2)],
            ...      "t3": [(3, 3)],
            ...      "t4": [(4, 4)],
            ... }

            Note that `t1` occurs twice on span `(1, 1)`, and that `t1` and `t2` have span `(1, 2)`
            in common.

            Suppose first that the sets `taxa_1` and `taxa_2` have no common taxa. In this
            case, the couples of spans on which to check the predicate are obtained by a simple
            cross-product, e.g.:

            >>> iterate_on_spans(spans, {"t1", "t2"}, {"t3", "t4"})
            ((1, 1), (3, 3))  # t1 (1) ⨉ t3
            ((1, 1), (3, 3))  # t1 (2) ⨉ t3
            ((1, 2), (3, 3))  # t1 (3) ⨉ t3
            ((1, 1), (4, 4))  # t1 (1) ⨉ t4
            ((1, 1), (4, 4))  # t1 (2) ⨉ t4
            ((1, 2), (4, 4))  # t1 (3) ⨉ t4
            ((2, 3), (3, 3))  # t2 (1) ⨉ t3
            ((1, 2), (3, 3))  # t2 (2) ⨉ t3
            ((2, 3), (4, 4))  # t2 (1) ⨉ t4
            ((1, 2), (4, 4))  # t2 (2) ⨉ t4

            Suppose now that the sets `taxa_1` and `taxa_2` have `t1` in common. In that case,
            we must suppress the diagonal of the cross-product, which not only carries zero useful
            information, but will make certain conditions fail. For instance, if we seek to include
            all the programs _not_ featuring two multiplications on the same line, the condition
            will be something like: “no multiplication is featured on the same line than another
            one”. Keeping the diagonal would add “or itself“ to the previous condition. Since any
            taxon is obviously featured on the same line than itself, no program would satisfy the
            condition.

            In the example below, note which lines are suppressed, and which are kept (particularly
            for `((1, 1), (1, 1))`):

            >>> iterate_on_spans(spans, {"t1", "t2"}, {"t1", "t3"})
            # ((1, 1), (1, 1))  # t1 (1) ⨉ t1 (1) (not generated)
              ((1, 1), (1, 1))  # t1 (1) ⨉ t1 (2)
              ((1, 1), (1, 2))  # t1 (1) ⨉ t1 (3)
              ((1, 1), (1, 1))  # t1 (2) ⨉ t1 (1)
            # ((1, 1), (1, 1))  # t1 (2) ⨉ t1 (2) (not generated)
              ((1, 1), (1, 2))  # t1 (2) ⨉ t1 (3)
              ((1, 2), (1, 1))  # t1 (3) ⨉ t1 (1)
              ((1, 2), (1, 1))  # t1 (3) ⨉ t1 (2)
            # ((1, 2), (1, 2))  # t1 (3) ⨉ t1 (3) (not generated)
              ((1, 1), (3, 3))  # t1 (1) ⨉ t3
              ((1, 1), (3, 3))  # t1 (2) ⨉ t3
              ((1, 2), (3, 3))  # t1 (3) ⨉ t3
              ((2, 3), (1, 1))  # t2 (1) ⨉ t1 (1)
              ((2, 3), (1, 1))  # t2 (1) ⨉ t1 (2)
              ((2, 3), (1, 2))  # t2 (1) ⨉ t1 (3)
              ((1, 2), (1, 1))  # t2 (2) ⨉ t1 (1)
              ((1, 2), (1, 1))  # t2 (2) ⨉ t1 (2)
              ((1, 2), (1, 2))  # t2 (2) ⨉ t1 (3)
              ((2, 3), (3, 3))  # t2 (1) x t3
              ((1, 2), (3, 3))  # t2 (2) x t3
        """
        for (taxon_1, taxon_2) in product(taxa_1, taxa_2):
            if taxon_1 in spans and taxon_2 in spans:
                if taxon_1 == taxon_2:  # exclude diagonal from iteration
                    yield from permutations(spans[taxon_1], 2)
                else:
                    yield from product(spans[taxon_1], spans[taxon_2])


# <graphviz>
# "__init__" -> "define_shortcuts"
# "__init__" -> "init_filter_state"
# "__init__" -> "add_imported_taxa"
# "init_filter_state" -> "taxa_of_programs"  [style=invis]
# "update_filter" -> "programs_and_taxa_of_patterns"
# "update_filter" -> "programs_or_taxa_of_patterns"
# "programs_and_taxa_of_patterns" -> "programs_of_pattern"
# "programs_and_taxa_of_patterns" -> "taxa_of_programs"
# "programs_and_taxa_of_patterns" -> "taxa_of_pattern"
# "programs_or_taxa_of_patterns" -> "programs_of_pattern"
# "programs_or_taxa_of_patterns" -> "taxa_of_pattern"
# "update_filter" -> "programs_of_criteria"
# "programs_of_criteria" -> "programs_of_pattern"
# "programs_of_criteria" -> "programs_of_taxa"
# "programs_of_criteria" -> "taxa_of_pattern"
# "programs_of_criteria" -> "programs_of_triple"
# "programs_of_triple" -> "taxa_of_pattern"
# "programs_of_triple" -> "programs_of_taxa"
# "programs_of_triple" -> "iterate_on_spans"
# "programs_of_criteria" -> "programs_of_negated_triple"
# "programs_of_negated_triple" -> "taxa_of_pattern"
# "programs_of_negated_triple" -> "programs_of_taxa"
# "programs_of_negated_triple" -> "iterate_on_spans"
# "update_filter" -> "impart_taxa"
# "update_filter" -> "include_programs"
# "update_filter" -> "exclude_programs"
# </graphviz>



r"""Search a program for the features specified in `spec.md`."""

from collections import defaultdict
from os.path import dirname
from pathlib import Path
from time import perf_counter
from typing import Callable, Dict, Iterator, List, Tuple

import regex  # type: ignore

from .derived_labels_db import DerivedLabelsDatabase
from .flatten_ast import ast, flatten_ast
from .goodies import print_fail
from .user_types import Label, LabelName, Labels, LabelsSpans, Program, Query, Source, Span

__pdoc__ = {
    "ProgramParser": "",
    "ProgramParser.__init__": True,
    "ProgramParser.__call__": True,
    "foobar": True,
}


def find_all_features(
    specifications: str,
    find_all: Callable = regex.compile(
        r"(?ms)"
        r"^\#{4}\s+Feature\s+`(.+?)`"  # capture the label's pattern
        r".+?\#{5}\s+Specification"  # ensure the sequel is in the Specification section
        r".+?```(.*?)\n+(.*?)\n?```"  # capture the language and the specification
    ).findall,
) -> Iterator[Tuple[LabelName, str, str]]:
    """Iterate on all tuples defining an algorithmic feature in the given specification text.

    Args:
        specifications (str): Normally, the contents of `spec.md`.
        find_all (Callable, optional): A function finding all feature-defining triples of the text.
            [Not to be explicitly provided.](developer_manual/index.html#default-argument-trick)

    Returns:
        Iterator[Tuple[LabelName, str, str]]: An iterator yielding all matching triples of the form:

            1. label name pattern (e.g., `"try"` or `"try_raise|try_except"`),
            2. language (currently, `"re"` or `"sql"`),
            3. specification (respectively, a regular expression pattern or an SQL query).
    """
    return find_all(specifications)


def pos_to_span(pos: List[str]) -> Span:
    """Convert a list of positions (as captured in a specification string) into a single span.

    Args:
        pos (List[str]): A non-empty list of strings of the form `f"{line_number}:{path}"`

    Returns:
        Span: A named tuple consisting of:

            - `start`: the line number extracted from the first string;
            - `end`: the line number extracted from the last string (the same as `start` if there
                is only one string);
            - `path`: the path extracted from the first string.

            Note that all other substrings, if any, are ignored.

    Examples:
        >>> pos_to_span(["1:path_1"])
        Span(1, 1, "path_1")
        >>> pos_to_span(["1:path_1", "2:path_2"])
        Span(1, 2, "path_1")
        >>> pos_to_span(["1:path_1", "2:path_2", "3:path_3", "4:path_4"])
        Span(1, 4, "path_1")
    """
    (start, path) = pos[0].split(":")
    (end, _) = pos[-1].split(":")
    return Span(int(start), int(end), path)


DEFAULT_SPEC_PATH = Path(dirname(__file__)) / "resources" / "spec.md"


def get_bindings(
    label_prefix: LabelName, captures: Dict[str, List[str]]
) -> Iterator[Tuple[LabelName, Span]]:
    r"""Analyze the matches of a feature, and bind the corresponding names and spans.

    Args:
        label_prefix (LabelName): The name of a feature specified by a regular expression. Note that
            such a name is literal: contrary to those defined by an SQL query, they cannot include
            the alternation meta-character (`"|"`).
        captures (Dict[str, List[str]]): A dictionary of the named groups and lists of all the
            captures of those groups, as matched by the function `regex.finditer()`. Contains a
            non-empty entry `"POS"` and, optionally, an entry `"SUFFIX"`. All other entries are
            ignored.

    Yields:
        Iterator[Tuple[LabelName, Span]]: At least one binding between a label name and a span.

    Examples:
        The simplest cases happen when there is one or several `"POS"`, but no `"SUFFIX"`:

        >>> get_bindings("null_operation", {"POS": ["4:1-3-1-5-1-"]})
        ("null_operation", Span(start=4, end=4, path="1-3-1-5-1-"))
        >>> get_bindings("while", {"POS": ["11:2-", "16:2-1-4-2-1-2-"]})
        ("while", Span(start=11, end=16, path="2-"))
        >>> get_bindings("divisibility_test", {
        ...     "POS": ["15:3-5-1-2-1-0-"],
        ...     "SUFFIX": []
        ... })
        ("divisibility_test", Span(start=15, end=15, path="3-5-1-2-1-0-"))

        When there are as many `"POS"` than `"SUFFIX"`, they are binded pairwise. The
        resulting spans cannot span multiple lines:

        >>> get_bindings("literal", {
        ...     "POS": ["16:3-5-1-2-1-1-1-0-"],
        ...     "SUFFIX": ["False"],
        ... })
        ("literal:False", Span(start=16, end=16, path="3-5-1-2-1-1-1-0-"))
        >>> get_bindings("if_test_atom", {
        ...     "POS": ["31:5-2-1-0-0-0-", "31:5-2-1-0-0-2-1-1-", "31:5-2-1-0-2-1-"],
        ...     "SUFFIX": ["b", "a", "greatest"],
        ... })
        ("if_test_atom:b", Span(start=31, end=31, path="5-2-1-0-0-0-"))
        ("if_test_atom:a", Span(start=31, end=31, path="5-2-1-0-0-2-1-1-"))
        ("if_test_atom:greatest", Span(start=31, end=31, path="5-2-1-0-2-1-"))

        In all other cases, `"POS"` is interpreted as delimiting only one occurrence (which can span
        multiple lines), and associated repeatedly with the different suffixes:

        >>> get_bindings("import_module", {"POS": ["1:1-"], "SUFFIX": ["m1", "m2", "m3"]})
        ("import_module:m1", Span(start=1, end=1, path="1-"))
        ("import_module:m2", Span(start=1, end=1, path="1-"))
        ("import_module:m3", Span(start=1, end=1, path="1-"))
        >>> get_bindings("function_decorator", {
        ...     "POS": ["1:1-", "1:1-2-1-", "2:1-2-2-", "3:1-2-3-", "5:1-5-1-"],
        ...     "SUFFIX": ["bizz", "foo", "bar"],
        ... })
        ("function_decorator:bizz", Span(start=1, end=5, path="1-")),
        ("function_decorator:foo", Span(start=1, end=5, path="1-")),
        ("function_decorator:bar", Span(start=1, end=5, path="1-")),

    Warning:
        Consider the problem of detecting the feature `for` in the following program:

        ```python
        for (i, j) in seq:
            pass
        ```

        With such a multi-line feature, it is normally enough to capture **two** positions, namely
        those of its start (1) and its end (2). However, in this case we want to suffix `for` with
        the names of the **two** iterations variables (`for:i` and `for:j`). If we devise the
        regular expression specification without special care:

        ```re
                ^(.*)/_type=For
        \n(?:\1.+\n)*?\1/_pos=(?P<POS>.+)
        (
        \n(?:\1.+\n)*?\1/target(/.+)?/id=(?P<SUFFIX>.+)
        )+
        \n(?:\1.+\n)* \1/.*/_pos=(?P<POS>.+)
        ```

        ... as many positions as there are suffixes will be passed to the present function:

        >>> get_bindings("for", {'POS': ['1:1-', '2:1-2-1-'], 'SUFFIX': ['i', 'j']})
        ("for:i", Span(start=1, end=1, path="1-"))
        ("for:j", Span(start=2, end=2, path="1-2-1-"))

        ... which will result in two (wrong) bindings instead of one. The workaround implemented in
        [spec.md](https://repo/paroxython/resources/spec.md#feature-for) consists in capturing as
        many supplementary positions as iteration variables:

        ```re
                ^(.*)/_type=For
        \n(?:\1.+\n)*?\1/_pos=(?P<POS>.+)
        (
        \n(?:\1.+\n)*?\1/(?P<_1>target(/.+)?)/_pos=(?P<POS>.+)
        \n(?:\1.+\n)*?\1/(?P=_1)             /id=(?P<SUFFIX>.+)
        )+
        \n(?:\1.+\n)* \1/.*/_pos=(?P<POS>.+)
        ```

        So, there are now two more positions in the captured dictionary. Since they are not at the
        ends of the list, they are ignored by `pos_to_span`, and the result is now correct:

        >>> get_bindings("for", {
        ...     'POS': ['1:1-', '1:1-0-0-1-', '1:1-0-0-2-', '2:1-2-1-'],
        ...      '_1': ['target/elts/1', 'target/elts/2'],  # ignored at this level
        ...      'SUFFIX': ['i', 'j']
        ... })
        ("for:i", Span(start=1, end=2, path="1-"))
        ("for:j", Span(start=1, end=2, path="1-"))
    """
    if captures.get("SUFFIX"):
        # There is a "SUFFIX" key and its value is not `[]`.
        if len(captures["POS"]) == len(captures["SUFFIX"]):
            # When there are as many matched positions as suffixes, associate them in parallel.
            for (suffix, pos) in zip(captures["SUFFIX"], captures["POS"]):
                yield (LabelName(f"{label_prefix}:{suffix}"), pos_to_span([pos]))
        else:
            # When there are more or less matched positions than suffixes, interpret the positions
            # as defining a unique span, and associate it with each suffix.
            span = pos_to_span(captures["POS"])
            for suffix in captures["SUFFIX"]:
                yield (LabelName(f"{label_prefix}:{suffix}"), span)
    else:
        # There is no "SUFFIX" key or its value is `[]`.
        yield (label_prefix, pos_to_span(captures["POS"]))


class ProgramParser:
    def __init__(self, spec_path: Path = DEFAULT_SPEC_PATH) -> None:
        """Collect the features specified at `spec_path`.

        Args:
            spec_path (Path, optional): The path of the specification file. Defaults to
                [spec.md](https://repo/paroxython/resources/spec.md).

        Raises:
            ValueError: Fail when a given feature appears multiple times in the specification file,
                and when the language is not `"re"` or `"sql"`, and the specification is non-empty.

        Description:
            Some features are searched for with a regular expression, some with an SQL query.
            Distribute them in their respective dictionary. Compile the regular expressions and
            pre-bind the appropriate method. Create an instance of the database of derived labels.
        """
        self.spec_path = spec_path
        text = self.spec_path.read_text()
        self.features: Dict[LabelName, regex.Pattern] = {}
        self.queries: Dict[LabelName, Query] = {}
        self.times: Dict[LabelName, float] = {LabelName("TOTAL"): 0.0}
        for (label_name, language, specification) in find_all_features(text):
            if label_name in self.features:  # pragma: no cover
                print_fail(f"Duplicated name '{label_name}'!")
            self.times[label_name] = 0.0
            if language == "re":
                self.features[label_name] = regex.compile(f"(?mx){specification}").finditer
            elif language == "sql":
                self.queries[label_name] = Query(specification)
            elif specification.strip() != "":  # pragma: no cover
                print_fail(f"Unknow language '{language}' for '{label_name}'!")
        self.derived_labels_database = DerivedLabelsDatabase()

    def __call__(self, program: Program, yield_failed_matches: bool = False) -> Labels:
        """Parse a given program and return its labels with their spans.

        Args:
            program (Program): A `Program` object, with its fields `source`, `addition` and
                `deletion` populated.
            yield_failed_matches (bool, optional): If `True`, add to the returned labels those
                which have not been found in the program. Each one is associated with an empty
                list of spans. For testing purposes only. Defaults to `False`.

        Returns:
            Labels: A sorted list of labels, associating each label name with its spans.

        Description:
            1. Parse the given program. If this fails or the resulting AST is empty, return a list
              consisting in a single label, `"ast_construction:"`, followed by the error name.
              Otherwise, flatten the AST.
            2. Search it for every feature which is specified by a regular expression. Take care to
              avoid adding those features which are scheduled for deletion at the same span.
            3. Update the matching features with those featured for addition.
            4. Use the features found so far to derive (directly and indirectly) all those specified
              by an SQL query.
            5. Return the found features.

        Note:
            The computation of the derived features is done in a single pass, which require them to
            be in a correct dependency order. This order is checked by `helpers/reformat_spec.py`,
            which is executed each time the tests are launched.
        """

        # If the program can be parsed and is nonempty, flatten its AST.
        try:
            tree = ast.parse(program.source)
        except (SyntaxError, ValueError) as exception:
            return [
                Label(
                    LabelName(f"ast_construction:{type(exception).__name__}"),
                    [Span(1, program.source.count("\n") + 1)],
                )
            ]
        # The next comment prevents a (wrong?) mypy error: "AST" has no attribute "body"
        if not tree.body:  # type: ignore
            return [Label(LabelName("ast_construction:EmptyProgramError"), [Span(0, 0)])]
        self.flat_ast = flatten_ast(tree)

        # Search the flat AST for every feature which is specified by a regular expression.
        labels: LabelsSpans = defaultdict(list)
        for (label_prefix, finditer) in self.features.items():
            start = perf_counter()
            failed_match = True
            for match in finditer(self.flat_ast, overlapped=True):
                failed_match = False
                for (label_name, span) in get_bindings(label_prefix, match.capturesdict()):
                    try:  # Unless this binding (without its path) is scheduled for deletion...
                        program.deletion[label_name].remove(Span(span.start, span.end))  # (no path)
                    except (KeyError, ValueError):  # ... actually bind the name with the span.
                        labels[label_name].append(span)
            elapsed = perf_counter() - start
            self.times[label_prefix] += elapsed
            self.times[LabelName("TOTAL")] += elapsed
            if yield_failed_matches and failed_match:
                labels[label_prefix] = []

        # Update the matching features with those featured for addition.
        for (label_name, spans) in program.addition.items():
            labels[label_name].extend(spans)
            labels[label_name].sort()

        # Now, use the features found so far to derive all those specified by an SQL query.
        result = [Label(*item) for item in labels.items()]
        self.derived_labels_database.create(result)  # The search is seeded with the know features.
        for (label_name, query) in self.queries.items():
            # Try to derive, from the DB current state, some label names matching `label_name`.
            start = perf_counter()
            derived_labels = self.derived_labels_database.read(query)
            elapsed = perf_counter() - start
            self.times[label_name] += elapsed
            self.times[LabelName("TOTAL")] += elapsed
            # Suppress the labels scheduled for deletion
            labels.clear()  # reuse the dictionary associating each label name to its spans
            for label in derived_labels:
                for span in label.spans:
                    try:  # Unless this binding (without its path) is scheduled for deletion...
                        program.deletion[label.name].remove(Span(span.start, span.end))
                    except (KeyError, ValueError):  # ... actually bind the name with the span.
                        labels[label.name].append(span)
            derived_labels = [Label(*item) for item in labels.items()]
            if derived_labels:
                # Bingo! update the DB state and the result with the new labels.
                self.derived_labels_database.update(derived_labels)
                result.extend(derived_labels)
            elif yield_failed_matches:
                result.append(Label(label_name, []))
        # Empty the DB to make it ready for the next program.
        self.derived_labels_database.delete()

        return sorted(result)

    def print_performances(self):  # pragma: no cover
        """Print a TSV report of elapsed times by features, sorted by decreasing time."""
        result = sorted(self.times.items(), key=lambda item: item[1], reverse=True)
        print()
        print("Elapsed times by features (in seconds)")
        print("--------------------------------------")
        for (name, seconds) in result:
            print(f"{seconds:8.4f}\t {name}")
        print()


if __name__ == "__main__":
    # Take an individual source, print its features and write its flat AST.
    from .goodies import couple_to_string

    path = Path("sandbox/source.py")
    source = path.read_text().strip()
    if source.startswith("1   "):
        source = regex.sub(r"(?m)^.{1,4}", "", source)
    for (i, line) in enumerate(source.split("\n"), 1):
        print(f"{i:<4}{line}")
    program = Program(source=Source(source), labels=[], taxa=[], addition={}, deletion={})
    print()
    parse = ProgramParser()
    acc = []
    for (name, spans) in sorted(parse(program, yield_failed_matches=False)):
        spans_as_string = ", ".join(map(couple_to_string, spans))
        acc.append(f"| `{name}` | {spans_as_string} |")
        # acc[-1] += " %s |" % ", ".join(f"{span.path}" for span in spans)
    print("\n".join(acc))
    Path("sandbox/flat_ast.txt").write_text(parse.flat_ast)



r"""
Read and execute a pipeline of commands and report the learning costs.

USAGE:
    ```
    paroxython recommend [options] DB_PATH
    ```

DESCRIPTION:
    ```plain
    Requires an existing database previously generated by the command:

        paroxython collect DIRECTORY

    ... and a pipeline of commands describing which concepts (taxa) are
    already known, which ones need to be illustrated or must be avoided. By
    default, it results in a list of recommended programs, along with their
    total learning costs and the individual learning cost of each of the
    concepts they feature. These results are presented in Markdown format.
    With the option `-o stdout` the program list is simply printed to stdout.
    ```

SHORTCUT:
    ```plain
    For DB_PATH, instead of the path of a tag database, you can provide the
    DIRECTORY previously collected. Paroxython will automatically look for the
    tag database at DIRECTORY/../DIRECTORY_db.json.
    ```

OPTIONS:
    ```plain
    -b --base=PATH      Value accessible, with the syntax "{base}", by any shell
                        command of the pipeline. If not specified, this is the
                        parent directory of DB_PATH. [default: ]
    -c --cost=STR       Learning cost assessment strategy. [default: zeno]
                        Currently available:
                        • zeno: the i-th edge of a taxon, if not already
                                imparted, costs 2^(-i). For instance, the taxon
                                old/new/new will costs 2^-2 + 2^-3 = 0.375.
                        • linear: simply count the number of new edges.
    -o --output=PATH    The path of the resulting report. If it is omitted, and
                        DB_PATH is of the form PREFIX_db.json, a value of
                        PREFIX_recommendations.md is used. If it is STDOUT, the
                        sorted list of recommended (but not hidden) programs, is
                        printed on the standard output. [default: ]
    -p --pipe=PATH      Path of the command pipeline. If it is omitted, and
                        DB_PATH is of the form PREFIX_db.json, a value of
                        PREFIX_pipe.py is used. If the associated file is
                        missing or malformed, no filter is applied. [default: ]
    -f --format=STR     The format of program titles in the report. This string
                        can contains the following identifiers (between braces):
                        • name: filename of the program.
                        • path: program key in the DB "programs" dictionary.
                        • prefix: name of DB_PATH, minus any "_db.json" suffix.
                        • absolute: absolute path to DB_PATH's parent folder.
                        • relative: relative path to DB_PATH's parent folder.
                        Shortcut: if format is "vscode", makes program names to
                            be clickable and opening in VS Code.
                        [default: `{name}`]
    ```

EXAMPLE:
    ```plain
    Use `-o stdout` to pipe the results into another shell command. For instance,
    let's assume that:

    - your program repository is `path/to/programs`;
    - your tag database is at its default location `path/to/programs_db.json`;
    - your pipeline is at its default location `path/to/programs_pipe.py`.

    Then, to copy (not move) all the recommended programs to an existing
    directory DEST, just do:

    paroxython recommend -o stdout path/to/programs_db.json \
        | xargs  -I {} sh -c "cp path/to/programs/{} DEST"
    ```
"""

import json
import sys
from pathlib import Path

import regex  # type: ignore
from typed_ast.ast3 import literal_eval

from .goodies import print_success, print_exit
from .recommend_programs import Recommendations


def cli_wrapper(args):
    db_path = Path(args["DB_PATH"])
    if not db_path.exists():
        print_exit(f"no file or directory at '{db_path.absolute()}': aborted.")
    parent_path = db_path.parent
    if db_path.is_dir():
        for suffix in ("_db", "-db"):
            db_path = parent_path / f"{db_path.name}{suffix}.json"
            if db_path.is_file():
                break
        else:
            print_exit(f"unable to locate a tag database in '{parent_path}': aborted.")
    m = regex.fullmatch(r"(.+[_-])db\.json", db_path.name)
    prefix = m[1] if m else ""
    pipeline_path = Path(args["--pipe"] or parent_path / f"{prefix}pipe.py")
    if pipeline_path.is_file():
        try:
            commands = literal_eval(pipeline_path.read_text())
        except Exception:  # Too many possible exceptions
            print_exit(f"the pipeline '{pipeline_path}' is malformed: aborted.")
    elif args["--pipe"]:
        print_exit(f"no pipeline at '{pipeline_path}': aborted.")
    else:
        commands = []
    stdout_backup = sys.stdout
    if args["--output"].upper() == "STDOUT":
        sys.stdout = sys.stderr
    title_format = args["--format"]
    if title_format.lower() == "vscode":
        title_format = "[`{name}`](vscode://file/{absolute}/{prefix}/{path})"
    title_format = title_format.format(
        name="{name}",
        path="{path}",
        prefix=prefix.rstrip("_-"),
        absolute=parent_path.resolve(),
        relative=parent_path,
    )
    rec = Recommendations(
        db=json.loads(db_path.read_text()),
        base_path=Path(args["--base"] or parent_path),
        assessment_strategy=args["--cost"],
        title_format=title_format,
    )
    rec.run_pipeline(commands)
    if args["--output"].upper() == "STDOUT":
        sys.stdout = stdout_backup
        return print("\n".join(sorted(rec.selected_programs - rec.hidden_programs)))
    output_path = Path(args["--output"] or parent_path / f"{prefix}recommendations.md")
    output_path.write_text(rec.get_markdown())
    print_success(f"Dumped: {output_path.resolve()}.\n")



"""
A wrapper around `paroxython.filter_programs.ProgramFilter`, feeding it with commands and assessing
the learning costs of its final state.

## Description

The recommendation system is initialized with a database created by `paroxython.make_db`. It parses
the given pipeline
([example](https://repo/examples/dummy/pipe.py)), checks
its commands and submits them to the filter, which in response updates these sets:

- `selected_programs`, the names of the recommended programs;
- `imparted_knowledge`, the taxa modelizing the imparted knowledge;
- `hidden_programs`/`hidden_taxa`, the programs/taxa to be removed from the final output.

When the pipeline is exhausted, it retrieves their values, compute the associated learning costs,
and produces a Markdown report
([example](https://repo/examples/simple/programs_recommendations.md))
whose each section consists in:

- the name of a recommended program;
- its total learning cost;
- its source code;
- the table of its taxa, lexicographically sorted, and accompanied by their spans and individual
  learning cost.

To maximize their utility, these sections are grouped by cost range, and each group is sorted by
increasing total cost and [SLOC](https://en.wikipedia.org/wiki/Source_lines_of_code).
"""

import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import regex  # type: ignore

from .assess_costs import LearningCostAssessor
from .filter_programs import ProgramFilter
from .goodies import (
    add_line_numbers,
    cost_bucket,
    couple_to_string,
    enumeration_to_txt_factory,
    print_warning,
    title_to_slug_factory,
)
from .user_types import (
    AssessedPrograms,
    Command,
    Criterion,
    JsonTagDatabase,
    Operation,
    ProgramPaths,
)

__pdoc__ = {
    "MalformedData": False,
    "Recommendations": "",
    "Recommendations.__init__": True,
}


class MalformedData(Exception):
    ...


class Recommendations:
    def __init__(
        self,
        db: JsonTagDatabase,
        base_path: Optional[Path] = None,
        title_format: str = "`{name}`",
        **kwargs,
    ) -> None:
        r"""Initialize a recommendation system for the given database of programs tagged with taxa.

        Args:
            db (JsonTagDatabase): A Python dictionary containing the JSON structure constructed by
                `paroxython.make_db.TagDatabase`, i.e. the result of the parsing, labelling and
                taxon-ifying of a set of programs
                ([example](https://repo/examples/simple/programs_db.json)).
                The execution relies on the fields `programs`, `taxa`, `importations` and
                `exportations`, but ignores the field `labels`.
            base_path (Optional[Path], optional): A path, accessible from any optional shell command
                under the name `"{base_path}"`. Defaults to `None`.
            title_format (str): a string containing a substring `{name}` to be interpolated in
                the final report with the relevant program's name. Defaults to `"\`{name}\`"`.
            **kwargs: May include the keyword argument `assessment_strategy`, transmitted to
                `paroxython.assess_costs.LearningCostAssessor`.

        Warning:
            `db` is modified by side-effect. After initialization, the taxa of some programs have
            been augmented with the taxa of those they import, associated with an empty list of
            spans. Exception: the taxa starting with `"meta/"` are not imported.
        """

        self.base_path = base_path
        self.title_format = title_format

        # copy locally some attributes and methods of a ProgramFilter instance
        program_filter = ProgramFilter(db)
        self.db_programs = program_filter.db_programs
        self.selected_programs = program_filter.selected_programs
        self.imparted_knowledge = program_filter.imparted_knowledge
        self.hidden_taxa = program_filter.hidden_taxa
        self.hidden_programs = program_filter.hidden_programs
        self.update_filter = program_filter.update_filter

        self.assess = LearningCostAssessor(self.db_programs, **kwargs)
        self.result: List[Tuple[int, Operation, ProgramPaths]] = []

    def run_pipeline(self, commands: Optional[List[Command]] = None) -> None:
        """Evolve the state of the filter accross pipeline commands.

        Description:
            Execute sequentially all the commands of the pipeline. For each command:

            1. Retrieve the operation (`"include"`, `"exclude"`, `"impart"` or `"hide"`, with an
              optional suffix `" any"` (implicit) or `" all"`).
            2. Retrieve the data by calling `Recommendations.parse_data`.
            3. Update the selected programs and/or impart the associated taxa and/or mark the
              programs or taxa as hidden.
            4. Log the newly filtered out programs in `self.result`.

            Finally, compute the learning costs of the remaining selected programs based on the
            imparted knowledge.

        Args:
            commands (Optional[List[Command]], optional): A list of dictionaries with two mandatory
                entries:

                - `"operation"`: a string among `"include"`, `"exclude"`, `"impart"` and `"hide"`,
                    with an optional suffix `" any"` (implicit) or `" all"`.
                - `"data"`: either a string consisting in a shell command, or an heterogeneous list
                    of _criteria_: patterns (matching either program paths or taxon names) or
                    semantic triples (of the form subject, predicate, object).

                Defaults to `None`, which is treated as an empty list.
        """
        commands = commands or []
        current = set(self.db_programs)
        print(f"\nProcessing {len(commands)} commands on {len(current)} programs.")

        # Execute sequentially all the commands of the pipeline
        for (i, command) in enumerate(commands, 1):

            # Retrieve the operation
            try:
                operation = command["operation"]
            except KeyError:
                print_warning(f"operation {i} is ignored (not specified).")
                continue
            (operation, n) = regex.subn(" all", "", operation)
            quantifier = "all" if n == 1 else "any"
            operation = Operation(operation.replace(" any", ""))
            if operation not in ("include", "exclude", "impart", "hide"):
                print_warning(f"operation {i} ({operation}) is ignored (unknown).")
                continue

            # Retrieve the data
            try:
                data = self.parse_data(command.get("data", []), operation)
            except MalformedData:
                print_warning(f"unable to interpret the data of operation {i} ({operation}).")
                continue
            if not data:
                print_warning(f"operation {i} ({operation}) is ignored (no data).")
                continue

            # Update the state of the filter: selected / hidden programs, imparted / hidden taxa
            self.update_filter(data, operation, quantifier)

            # Log the statistics of the filter state for the last operation
            (previous, current) = (current, set(self.selected_programs))
            self.result.append((i, operation, sorted(previous - current)))
            print(f"  {len(current)} programs remaining after operation {i} ({operation}).")

        self.assess.set_imparted_knowledge(self.imparted_knowledge)
        self.assessed_programs = self.assess(self.selected_programs)

    def parse_data(
        self,
        data: Union[str, List[Criterion]],
        operation: Operation,
    ) -> List[Criterion]:
        """Retrieve the list of criteria to which to apply the given operation.

        Description:
            In case `data` is a list of criteria, return it unchanged. Otherwise, if it is a string,
            execute it as a shell command, capture its output and return it as a list of strings.
            Otherwise, fail.

        Note:
            A shell command cannot produce semantic triples. Its output is necessarily interpreted
            as a list of patterns.

        Args:
            data (Union[str, List[Criterion]]): Either a shell command (`str`) or a list of
                criteria. A criterion is either a pattern (`str`) or a semantic triple
                (`Tuple[str, str, str]`) subject, predicate, object.
            operation (Operation): Either `"include"`, `"exclude"`, `"impart"` or `"hide"`
                (note that the optional suffix `" all"` has been previously suppressed).

        Raises:
            MalformedData: Indicates that the field `"data"` of a pipeline command cannot
                be interpreted.

        Returns:
            List[Criterion]: A list of criteria.
        """
        if isinstance(data, list):  # The object can either be a list of criteria...
            return data
        elif isinstance(data, str):  # ... or a shell command printing them on stdout
            criteria = subprocess.run(
                str(data).format(base_path=self.base_path),  # str() needed by mypy
                stdout=subprocess.PIPE,  # From Python 3.7, these two arguments can be
                stderr=subprocess.PIPE,  # ... replaced by: capture_output=True
                encoding="utf-8",
                shell=True,
                check=True,
            )
            return [criterion for criterion in criteria.stdout.strip().split("\n")]
        else:
            raise MalformedData

    def get_markdown(
        self,
        span_column_width: int = 30,
        sorting_strategy: str = "by_cost_and_sloc",
        grouping_strategy: str = "by_cost_bucket",
    ) -> str:
        """Iterate on the results. Group, sort and output them."""

        # Define some helper functions.

        title_to_slug = title_to_slug_factory()
        spans_to_html = enumeration_to_txt_factory(span_column_width, "_imported_")
        display_count = lambda n: f"{n} program" + ("" if n == 1 else "s")

        sorting_key = None
        if sorting_strategy == "by_cost_and_sloc":
            sorting_key = lambda x: (x[0], len(self.db_programs[x[1]]["source"].split("\n")))
        elif sorting_strategy == "lexicographic":
            sorting_key = lambda x: x[1]

        grouping_key = lambda x: "(default: no group)"
        if grouping_strategy == "by_cost_bucket":
            grouping_key = cost_bucket

        # Group resulting programs into cost buckets, and sort each group by increasing difficulty.

        toc_data: Dict[str, AssessedPrograms] = defaultdict(list)
        for (cost, program_path) in self.assessed_programs:
            if program_path in self.hidden_programs:
                continue
            toc_data[grouping_key(cost)].append((cost, program_path))
        for costs_and_program_paths in toc_data.values():
            costs_and_program_paths.sort(key=sorting_key)

        # Accumulate simultaneously the TOC and the contents.

        toc: List[str] = ["# Table of contents"]
        contents: List[str] = ["# Recommended programs"]

        for (bounds, costs_and_program_paths) in toc_data.items():

            title = f"{display_count(len(costs_and_program_paths))} of learning cost {bounds}"
            toc.append(f"- [`{title}`](#{title_to_slug(title)})")
            contents.append(f"\n## {title}")

            for (cost, program_path) in costs_and_program_paths:
                program_info = self.db_programs[program_path]
                name = regex.sub(r"(?:.+/)?(.+)", r"\1", program_path)  # last segment of the path
                program_title = self.title_format.format(path=program_path, name=name)
                title = f"Program {program_title} (learning cost {cost})"
                slug = title_to_slug(f"Program {name} (learning cost {cost})")
                toc.append(f"    - [`{name}`](#{slug})")
                contents.append(f"\n### {title}")
                contents.append(f"\n```python\n{add_line_numbers(program_info['source'])}\n```")
                contents.append("\n| Cost  | Taxon | Location |")
                contents.append("|" + "----|" * 3)
                items = sorted(
                    program_info["taxa"].items(),
                    key=lambda x: f"~{x[0]}" if x[0].startswith("meta/") else x[0],
                )
                for (taxon_name, spans) in items:
                    if taxon_name in self.hidden_taxa:
                        continue
                    taxon_cost = self.assess.taxon_cost(taxon_name)
                    s = spans_to_html(", ".join(map(couple_to_string, spans)))
                    contents.append(f"| {taxon_cost} | `{taxon_name}` | {s} |")
                contents.append("\n---")

        def programs_to_html(description: str, programs: ProgramPaths) -> str:
            details = []
            if programs:
                description = description.replace(" 1 programs", " 1 program")
                details = ["<ol>"]
                for program in sorted(programs):
                    details.append(f"  <li><code>{program}</code></li>")
                details.append("</ol>")
            return (
                "<details>\n"
                + f"  <summary>{description}.</summary>\n  "
                + "\n  ".join(details)
                + "\n</details>\n"
            )

        n = len(self.db_programs)
        summary: List[str] = ["\n# Summary"]
        summary.append(programs_to_html(f"{n} initially", list(self.db_programs)))
        for (i, operation, removed_programs) in self.result:
            n -= len(removed_programs)
            title = f"{n} remaining after operation {i} ({operation}) has filtered out {len(removed_programs)} programs"
            summary.append(programs_to_html(title, removed_programs))

        return "\n".join(toc + contents + summary)


if __name__ == "__main__":
    ast = __import__("ast")
    json = __import__("json")
    path = Path("../algo/programs")
    rec = Recommendations(
        db=json.loads(Path(f"{path}_db.json").read_text()),
        base_path=path.parent,
    )
    rec.run_pipeline(ast.literal_eval(Path(f"{path}_pipe.py").read_text()))
    output_path = Path(f"{path}_recommendations.md")
    output_path.write_text(rec.get_markdown())
    print(f"Dumped: {output_path.resolve()}.\n")



"""An in-memory database collecting the labels of a given program.

### Main table

The database is not really relational. It merely consists in a table `t` where each row is a
distinct occurrence of every label featured by a given program. Initially, `t` is populated with
the labels found by the queries specified by a regular expression of `spec.md`.

The SQL queries are then executed, and the resulting labels (known as _derived_) inserted into `t`
as they are found.

Note that the results of a given SQL query may themselves be required by a subsequent SQL query.

#### Example

The following simple SQL query of `spec.md` finds the loops of a program. The resulting label names
(e.g., `"loop:while"`) consist in a fixed prefix `"loop"` and a suffix denoting the category of
the loop (either `"for"` or `"while"`).

```sql
SELECT "loop",
    name_prefix,
    span,
    path
FROM t
WHERE name_prefix IN ("for", "while")
GROUP BY path
```

The execution of this query comes after the evaluation of all regular expressions, in particular
those collecting the features `"for"` and `"while"`. It simply gathers them under the prefix
`"loop"`, and qualifies them by their category.

The resulting labels will later be used by other SQL queries to search for features such as
`"loop_with_break"`, `"loop_with_late_exit"`, `"loop_with_else"`, and so on.

### Subtables

For simplicity and performance reasons, a bunch of subtables are extracted on the fly from `t`.
Actually, every time a query relies on a non-existing table named `t_PREFIX`, this table is
dynamically created as the set of rows of `t` whose value on column `name_prefix` is `PREFIX`.

#### Example

The following SQL query of `spec.md` relies on the previous one to find the loops featuring a
`break` statement. It is more complicated, since it suffixes `"loop_with_break"` with the span of
the smallest enclosing loop. Anyway, it relies on the results of both the previous SQL query and a
trivial regular expression matching the `break` statements.

```sql
SELECT "loop_with_break",
    l.name_suffix,
    max(l.span_start) || "-" || min(l.span_end),
    max(l.path)
FROM t_loop l
JOIN t_break b ON (b.path GLOB l.path || "*-")
GROUP BY b.rowid
```

Instead of referring the corresponding rows in `t` (which would complicate the query), two
subtables restricted to them are referred, `t_loop` and `t_break`. For example, if a `t_loop`
subtable does not exist yet, the system extracts it from `t` by executing:

```sql
CREATE TABLE t_loop AS
SELECT *
FROM t
WHERE name_prefix = 'loop'
```

The newly created table is then available for all subsequent queries that may need it, which can
significantly speed up their execution.

### Note

Using views instead of tables turns out to be slightly slower. Also, for whatever mysterious
reason, several queries produce an incorrect result.
"""

import sqlite3
from collections import defaultdict
from typing import Any, Set, Callable, Dict

import regex  # type: ignore

from .goodies import couple_to_string, print_warning
from .user_types import Label, LabelName, Labels, Query, Span


class DerivedLabelsDatabase:

    columns = (
        # use rowid as primary key:
        "name TEXT",
        "name_prefix TEXT",
        "name_suffix TEXT",
        "span TEXT",
        "span_start INT",
        "span_end INT",
        "path TEXT",
    )
    """The structure of the main table `t`, consisting of the following columns:

    Column        | Type   | Description and example
    :-------------|:-------|:----------------------------------------------
    `rowid`       | `INT`  | Primary key (automatically generated)
    `name`        | `TEXT` | Complete name of the label, e.g.
                  |        | `"loop_with_early_exit:for:break"`
    `name_prefix` | `TEXT` | Its prefix (part before the first colon, here
                  |        | `"loop_with_early_exit"`
    `name_suffix` | `TEXT` | Its suffix (may contain one or more colons),
                  |        | here `"for:break"`
    `span`        | `TEXT` | Numbers of the first and last lines of the
                  |        | feature as a hyphen-separated string, e.g. `"2-8"`
    `span_start`  | `INT`  | Its first line number, here `2`
    `span_end`    | `INT`  | Its first line number, here `8`
    `path`        | `TEXT` | String encoding the beginning of the feature
                  |        | as its path from the root of the AST, e.g.
                  |        | `"3-5-3-0-1-2-2-1-"`
    """

    def __init__(self):
        self.c = sqlite3.connect(":memory:")
        self.c.create_function("regexp", narg=2, func=lambda rex, s: bool(regex.match(rex, s)))
        # References:
        # - https://www.sqlite.org/inmemorydb.html
        # - https://www.sqlite.org/lang_expr.html#regexp
        # - https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection.create_function

    def create(
        self,
        labels: Labels,
        create_main_table: str = f"CREATE TABLE t ({','.join(columns)})",  # never provided
    ) -> None:
        """Create the main table and populate it with regex-specified labels.

        Args:
            labels (Labels): The label names and spans featured by a certain program. Only those
                specified by a regular expression in `spec.md` are included. It is the very purpose
                of this module to add the remaining ones, specified by a SQL query.
            create_main_table (str, optional): The SQL query creating the table. Not to be
                explicitly provided. Defaults to `f"CREATE TABLE t ({','.join(columns)})"`.
        """
        self.c.execute(create_main_table)
        self.update(labels)
        self.subtables: Set[LabelName] = set()

    def read(
        self,
        query: Query,
        prerequisites: Callable = regex.compile(r"(?m)\b(?:FROM|JOIN) t_(\w+)").findall,
        create_subtable: str = "CREATE TABLE t_{0} AS SELECT * FROM t WHERE name_prefix = '{0}'",
    ) -> Labels:
        r"""Apply the given SQL query to the currently known labels, and returns the new labels.

        Args:
            query (Query): The SQL specification of a certain feature, as defined in `spec.md`.
            prerequisites (Callable, optional): A function taking an SQL query, and returning the
                table names referred to in the FROM clause.
                [Not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
                Defaults to `regex.compile(r"(?m)\b(?:FROM|JOIN) t_(\w+)").findall`.
            create_subtable (str, optional): The SQL query creating a subtable from a label prefix.
                Not to be explicitly provided.
                Defaults to `"CREATE TABLE t_{0} AS SELECT * FROM t WHERE name_prefix = '{0}'"`.

        Returns:
            Labels: The additional labels found by the given SQL query.

        Note:
            As a side-effect, create in the database any non-existing subtable of `t` referred to
            as a prerequisite.
        """
        # Parse the prerequisites of the query, and create the corresponding subtables if needed.
        for label_name in prerequisites(query):
            if label_name not in self.subtables:
                self.c.execute(create_subtable.format(label_name))
            self.subtables.add(label_name)
        self.c.commit()
        # Execute the query and return the resulting labels, deduplicated by (label_name, path).
        labels_spans: Dict[LabelName, Dict[Span, Any]] = defaultdict(dict)
        try:
            row_iterator = self.c.execute(query)
        except Exception as exception:  # pragma: no cover
            print_warning(f"problem in the following query:\n\n{query}\n")
            raise exception
        for (name_prefix, name_suffix, span_string, path) in row_iterator:
            label_name = f"{name_prefix}:{name_suffix}" if name_suffix != "" else name_prefix
            span = span_string.split("-")
            labels_spans[label_name][Span(int(span[0]), int(span[-1]), path)] = None
        return [Label(name, list(spans)) for (name, spans) in labels_spans.items()]

    def update(
        self,
        labels: Labels,
        update_query: str = f"INSERT INTO t VALUES ({','.join('?' * len(columns))})",
    ) -> None:
        """Insert a set of labels into the main table `t`.

        Args:
            labels (Labels): The labels to be inserted.
            update_query (str, optional): The SQL query inserting a row in `t`. Not to be
                explicitly provided. Defaults to `f"INSERT INTO t VALUES ({','.join('?' *
                len(columns))})"`.
        """
        rows = []
        for (name, spans) in labels:
            for span in spans:
                (prefix, _, suffix) = name.partition(":")  # split on the first colon
                span_string = couple_to_string(span)
                rows.append((name, prefix, suffix, span_string, span.start, span.end, span.path))
        self.c.executemany(update_query, rows)

    def delete(self) -> None:
        """Empty the database after having executed all the SQL queries."""
        self.c.execute("DROP TABLE t")
        for label_name in self.subtables:
            self.c.execute(f"DROP TABLE t_{label_name}")

    def __str__(self) -> str:  # pragma: no cover
        rows = ["name name_prefix name_suffix span span_start span_end path".split()]
        rows.extend(sorted(list(map(str, row)) for row in self.c.execute("SELECT * FROM t")))
        widths = list(map(len, [max(column, key=len) for column in zip(*rows)]))
        result = [""]
        for row in rows:
            result.append(" | ".join(s + " " * (w - len(s)) for (w, s) in zip(widths, row)))
        result[1:1] = ["-+-".join("-" * w for w in widths)]
        return "\n".join(result)



#! /usr/bin/env python
"""Command line interface for the main functionalities of Paroxython.

USAGE:
    ```
    paroxython [--version] [--help] COMMAND [ARGS...]
    ```

COMMANDS:
    ```plain
    tag         Tag one Python file and output the table of its taxa or
                labels.
    collect     Walk a directory, tag its Python files and make a JSON
                database of the results.
    recommend   Read and execute a pipeline of commands and report the learning
                costs.
    ```

Type `paroxython COMMAND --help` for more information on a specific command.
"""
import sys
from importlib import import_module

import regex  # type: ignore
from docopt import docopt  # type: ignore

from . import PAROXYTHON_VERSION
from .goodies import print_exit

if sys.version < "3.6":  # pragma: no cover
    print_exit(f"Paroxython requires Python 3.6 or later to run.\nThis version is {sys.version}.")


def main():
    doc = regex.sub(r"(?m)^ *```.*\n", "", __doc__)  # suppress Markdown code delimiters
    args = docopt(doc, version=f"paroxython version {PAROXYTHON_VERSION}", options_first=True)
    if not args.get("COMMAND"):  # pragma: no cover
        print_exit(doc)
    command_name = args["COMMAND"].lower()
    command_args = [command_name] + args["ARGS"]
    if command_name not in ("tag", "recommend", "collect"):
        print_exit(f"{command_name} is not a valid command. Type 'paroxython --help'.")
    command = import_module(f"paroxython.cli_{command_name}", "cli_wrapper")
    command.__doc__ = regex.sub(r"(?m)^ *```.*\n", "", command.__doc__)
    command.cli_wrapper(docopt(command.__doc__, argv=command_args))


if __name__ == "__main__":
    main()
